#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from type_inference.type_inference_service import TypeInference
from type_inference.types.edge import Equality, EqualityOfElement, FieldBelonging, PredicateArgument
from type_inference.types.expression import Variable, PredicateAddressing, NumberLiteral, SubscriptAddressing, \
  StringLiteral, RecordLiteral, ListLiteral
from type_inference.types.types_graph import TypesGraph
from type_inference.types.variable_types import NumberType, StringType, ListType, RecordType, AnyType, BoolType

number = NumberType()
string = StringType()


class TestTypeInference(unittest.TestCase):
  def test_when_num(self):
    # 'Q(x) :- x == 1'
    graph = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    x_var = Variable('x')
    graph.Connect(Equality(q_col0, x_var, (0, 0)))
    graph.Connect(Equality(x_var, NumberLiteral(), (0, 0)))
    graphs = dict()
    graphs['Q'] = graph

    TypeInference(graphs).Infer()

    self.assertEqual(q_col0.type, number)
    self.assertEqual(x_var.type, number)

  def test_when_inclusion(self):
    # Q(x) :- x in Range(10)
    graph = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    x_var = Variable('x')
    graph.Connect(Equality(q_col0, x_var, (0, 0)))
    graph.Connect(EqualityOfElement(PredicateAddressing('Range', 'logica_value'), x_var, (0, 0)))
    graph.Connect(Equality(PredicateAddressing('Range', 'col0'), NumberLiteral(), (0, 0)))
    graphs = dict()
    graphs['Q'] = graph

    TypeInference(graphs).Infer()

    self.assertEqual(q_col0.type, number)
    self.assertEqual(x_var.type, number)

  def test_when_list(self):
    # Q(Range(10));
    graph = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    graph.Connect(Equality(PredicateAddressing('Range', 'logica_value'), q_col0, (0, 0)))
    graph.Connect(Equality(PredicateAddressing('Range', 'col0'), NumberLiteral(), (0, 0)))
    graphs = dict()
    graphs['Q'] = graph

    TypeInference(graphs).Infer()

    self.assertEqual(q_col0.type, ListType(number))

  def test_when_opened_record(self):
    # Q(x) :- x.a == 1, x.b == "string"
    graph = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    x_var = Variable('x')
    a = SubscriptAddressing(x_var, 'a')
    b = SubscriptAddressing(x_var, 'b')
    graph.Connect(Equality(q_col0, x_var, (0, 0)))
    graph.Connect(FieldBelonging(x_var, a, (0, 0)))
    graph.Connect(FieldBelonging(x_var, b, (0, 0)))
    graph.Connect(Equality(a, NumberLiteral(), (0, 0)))
    graph.Connect(Equality(b, StringLiteral(), (0, 0)))
    graphs = dict()
    graphs['Q'] = graph

    TypeInference(graphs).Infer()

    self.assertEqual(q_col0.type, RecordType({'a': number, 'b': string}, True))

  def test_when_opened_record_with_opened_record(self):
    # Q(x) :- x.a == 1, x.b.c == "string"
    graph = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    x_var = Variable('x')
    a = SubscriptAddressing(x_var, 'a')
    b = SubscriptAddressing(x_var, 'b')
    c = SubscriptAddressing(b, 'c')
    graph.Connect(Equality(q_col0, x_var, (0, 0)))
    graph.Connect(FieldBelonging(x_var, a, (0, 0)))
    graph.Connect(FieldBelonging(x_var, b, (0, 0)))
    graph.Connect(FieldBelonging(b, c, (0, 0)))
    graph.Connect(Equality(a, NumberLiteral(), (0, 0)))
    graph.Connect(Equality(c, StringLiteral(), (0, 0)))
    graphs = dict()
    graphs['Q'] = graph

    TypeInference(graphs).Infer()

    self.assertEqual(q_col0.type, RecordType({'a': number, 'b': RecordType({'c': string}, True)}, True))

  def test_when_linked_with_known_predicates(self):
    # Q(x, y) :- T(x), Num(x), y == F(x);
    # T(x) :- x == 10;
    # F(x) :- x;
    graph_q = TypesGraph()
    x_var_q = Variable('x')
    y_var_q = Variable('y')
    graph_q.Connect(Equality(PredicateAddressing('Q', 'col0'), x_var_q, (0, 0)))
    graph_q.Connect(Equality(PredicateAddressing('Q', 'col1'), y_var_q, (0, 0)))
    graph_q.Connect(Equality(x_var_q, PredicateAddressing('T', 'col0'), (0, 0)))
    graph_q.Connect(Equality(x_var_q, PredicateAddressing('F', 'col0'), (0, 0)))
    graph_q.Connect(Equality(x_var_q, y_var_q, (0, 0)))
    graph_q.Connect(Equality(x_var_q, PredicateAddressing('Num', 'col0'), (0, 0)))
    graphs = dict()
    graphs['Q'] = graph_q

    graph_t = TypesGraph()
    x_var_t = Variable('x')
    graph_t.Connect(Equality(PredicateAddressing('T', 'col0'), x_var_t, (0, 0)))
    graph_t.Connect(Equality(x_var_t, NumberLiteral(), (0, 0)))
    graphs['T'] = graph_t

    graph_f = TypesGraph()
    x_var_f = Variable('x')
    graph_f.Connect(Equality(PredicateAddressing('F', 'col0'), x_var_f, (0, 0)))
    graphs['F'] = graph_f

    TypeInference(graphs).Infer()

    self.assertEqual(x_var_t.type, number)
    self.assertEqual(y_var_q.type, number)

  def test_when_field_is_function_result(self):
    # Q(x) :- x.a == T(), x.b == "string"
    # T() = 7
    graph_q = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    x_var = Variable('x')
    a = SubscriptAddressing(x_var, 'a')
    b = SubscriptAddressing(x_var, 'b')
    graph_q.Connect(Equality(q_col0, x_var, (0, 0)))
    graph_q.Connect(FieldBelonging(x_var, a, (0, 0)))
    graph_q.Connect(FieldBelonging(x_var, b, (0, 0)))
    graph_q.Connect(Equality(a, PredicateAddressing('T', 'logica_value'), (0, 0)))
    graph_q.Connect(Equality(b, StringLiteral(), (0, 0)))
    graphs = dict()
    graphs['Q'] = graph_q

    graph_t = TypesGraph()
    graph_t.Connect(Equality(PredicateAddressing('T', 'logica_value'), NumberLiteral(), (0, 0)))
    graphs['T'] = graph_t

    TypeInference(graphs).Infer()

    self.assertEqual(q_col0.type, RecordType({'a': number, 'b': string}, True))

  def test_when_plus_operator(self):
    # Q(x + y) :- T(x), T(y);
    graph_q = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    x_var_q = Variable('x')
    y_var_q = Variable('y')
    t_col0 = PredicateAddressing('T', 'col0')
    graph_q.Connect(Equality(q_col0, PredicateAddressing('+', 'logica_value'), (0, 0)))
    graph_q.Connect(Equality(x_var_q, PredicateAddressing('+', 'left'), (0, 0)))
    graph_q.Connect(Equality(y_var_q, PredicateAddressing('+', 'right'), (0, 0)))
    graph_q.Connect(Equality(x_var_q, t_col0, (0, 0)))
    graph_q.Connect(Equality(y_var_q, t_col0, (0, 0)))
    graphs = dict()
    graphs['Q'] = graph_q

    graph_t = TypesGraph()
    x_var_t = Variable('x')
    graph_t.Connect(Equality(PredicateAddressing('T', 'col0'), x_var_t, (0, 0)))
    graph_t.Connect(Equality(x_var_t, NumberLiteral(), (0, 0)))
    graphs['T'] = graph_t

    inferred_rules = TypeInference(graphs).Infer()

    self.assertEqual(q_col0.type, number)
    self.assertEqual(x_var_t.type, number)

  def test_when_opened_record_with_list(self):
    # Q(x, i) :- x.a == "range", x.b == Range(10), i in x.b
    graph = TypesGraph()
    graphs = dict()
    graphs['Q'] = graph
    x_var = Variable('x')
    i_var = Variable('i')
    a = SubscriptAddressing(x_var, 'a')
    b = SubscriptAddressing(x_var, 'b')
    graph.Connect(Equality(PredicateAddressing('Q', 'col0'), x_var, (0, 0)))
    graph.Connect(Equality(Variable('col1'), i_var, (0, 0)))
    graph.Connect(FieldBelonging(x_var, a, (0, 0)))
    graph.Connect(FieldBelonging(x_var, b, (0, 0)))
    graph.Connect(Equality(a, StringLiteral(), (0, 0)))
    graph.Connect(Equality(b, PredicateAddressing('Range', 'logica_value'), (0, 0)))
    graph.Connect(EqualityOfElement(b, i_var, (0, 0)))

    TypeInference(graphs).Infer()

    self.assertEqual(x_var.type, RecordType({'a': string, 'b': ListType(number)}, True))
    self.assertEqual(i_var.type, number)

  def test_when_closed_record(self):
    # Q({a: 1, b: "string"})
    graph = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    graph.Connect(Equality(q_col0, RecordLiteral({'a': NumberLiteral(), 'b': StringLiteral()}), (0, 0)))
    graphs = dict()
    graphs['Q'] = graph

    TypeInference(graphs).Infer()

    self.assertEqual(q_col0.type, RecordType({'a': number, 'b': string}, False))
    self.assertEqual(q_col0.type.is_opened, False)

  def test_when_closed_record_with_closed_record_with_list(self):
    # Q({a: 1, b: {c: []}})
    graph = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    inner_record = RecordLiteral({'c': ListLiteral([])})
    graph.Connect(Equality(q_col0, RecordLiteral({'a': NumberLiteral(), 'b': inner_record}), (0, 0)))
    graphs = dict()
    graphs['Q'] = graph

    TypeInference(graphs).Infer()

    expected = RecordType({'a': number, 'b': RecordType({'c': ListType(AnyType())}, False)}, False)
    self.assertEqual(q_col0.type, expected)

  def test_when_single_compare_operator(self):
    # Q(x < 2) :- x == 1;
    graph = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    x_var = Variable('x')
    graph.Connect(Equality(q_col0, PredicateAddressing('<', 'logica_value'), (0, 0)))
    graph.Connect(Equality(PredicateAddressing('<', 'left'), x_var, (0, 0)))
    graph.Connect(Equality(PredicateAddressing('<', 'right'), NumberLiteral(), (0, 0)))
    graph.Connect(Equality(x_var, NumberLiteral(), (0, 0)))
    graphs = dict()
    graphs['Q'] = graph

    TypeInference(graphs).Infer()

    self.assertEqual(q_col0.type, BoolType())

  def test_when_two_compare_operators_with_different_types(self):
    # Q("h" < "a", 1 < 2);
    graph = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    q_col1 = PredicateAddressing('Q', 'col1')
    graph.Connect(Equality(q_col0, PredicateAddressing('<', 'logica_value'), (0, 0)))
    graph.Connect(Equality(PredicateAddressing('<', 'left'), StringLiteral(), (0, 0)))
    graph.Connect(Equality(PredicateAddressing('<', 'right'), StringLiteral(), (0, 0)))
    graph.Connect(Equality(q_col1, PredicateAddressing('<', 'logica_value', 1), (0, 0)))
    graph.Connect(Equality(PredicateAddressing('<', 'left', 1), NumberLiteral(), (0, 0)))
    graph.Connect(Equality(PredicateAddressing('<', 'right', 1), NumberLiteral(), (0, 0)))
    graphs = dict()
    graphs['Q'] = graph

    TypeInference(graphs).Infer()

    self.assertEqual(q_col0.type, BoolType())
    self.assertEqual(q_col1.type, BoolType())

  def test_when_compare_records_fields(self):
    # Q(y) :- x == {b: 5}, y in Range(10), y < x.b;
    graph = TypesGraph()
    q_col0 = PredicateAddressing('Q', 'col0')
    y_var = Variable('y')
    x_var = Variable('x')
    b = SubscriptAddressing(x_var, 'b')
    range_lv = PredicateAddressing('Range', 'logica_value')
    range_col0 = PredicateAddressing('Range', 'col0')
    graph.Connect(Equality(q_col0, y_var, (0, 0)))
    graph.Connect(FieldBelonging(x_var, b, (0, 0)))
    graph.Connect(Equality(x_var, RecordLiteral({'b': NumberLiteral()}), (0, 0)))
    graph.Connect(Equality(PredicateAddressing('<', 'left'), y_var, (0, 0)))
    graph.Connect(Equality(PredicateAddressing('<', 'right'), b, (0, 0)))
    graph.Connect(EqualityOfElement(range_lv, y_var, (0, 0)))
    graph.Connect(PredicateArgument(range_lv, range_col0, (0, 0)))
    graphs = dict()
    graphs['Q'] = graph

    TypeInference(graphs).Infer()

    self.assertEqual(q_col0.type, number)



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Library to draw ASCII directed acyclic graph.


class Graph(object):
  def __init__(self, nodes, edges):
    self.nodes = nodes
    self.edges = edges
    self.dependencies = {n: [] for n in nodes}
    self.dependants = {n: [] for n in nodes}
    for a, b in edges:
      self.dependencies[b] += [a]
      self.dependants[a] += [b]
    self.SortNodes()

  def SortNodes(self):
    placed = set()
    result = []
    while len(placed) < len(self.nodes):
      for node in self.nodes:
        if node in placed:
          continue
        for d in self.dependencies[node]:
          if d not in placed:
            break
        else:
          result.append(node)
          placed |= {node}
    self.nodes = result

  def SimulateDelivery(self):
    all_tracks = set(range(len(self.edges)))
    tracks_image = []
    carts = {}
    busy_tracks = set()
    get_track = lambda: min(all_tracks - busy_tracks)
    for node in self.nodes:
      new_tracks = set()
      done_tracks = set()
      num_used_tracks = lambda: (
        0 if not busy_tracks else max(busy_tracks) + 1)
      dependants = set(self.dependants[node])

      # Adding carts towards all dependants.
      # TODO: Write this to be linear in len(self.dependants[node]).
      # We want to traverse over dependants in the order in which they occur
      # in self.nodes.
      for d in self.nodes:
        if d in dependants:
          track = get_track()
          carts[node, d] = track
          busy_tracks |= {track}
          new_tracks |= {track}

      # Retracting carts arriving from dependencies.
      plot_tracks = num_used_tracks()
      for d in self.dependencies[node]:
        track = carts[d, node]
        del carts[d, node]
        busy_tracks -= {track}
        done_tracks |= {track}

      # Drawing a row.
      row = ''
      char_space = ' '
      char_start = '╥'
      char_end = '╨'
      char_pipe = '║'
      for i in range(plot_tracks):
        if i in new_tracks:
          row += char_start
        elif i in done_tracks:
          row += char_end
        elif i in busy_tracks:
          row += char_pipe
        else:
          row += char_space

      # Appendig the row.
      tracks_image.append(row)
      tracks_image.append(''.join(
          (char_pipe if c == char_start or c == char_pipe else char_space
           for c in row)))
    width = max(map(len, tracks_image))
    tracks_image = [row + char_space * (width - len(row))
                    for row in tracks_image]
    char_field = '▚'
    field = f' {char_field}  '
    tracks_image = [
      row + field + str(self.nodes[i // 2])
      if i % 2 == 0 else row + field
      for i, row in enumerate(tracks_image)]
    # Producing a condenced 
    return tracks_image[::2]

  def GetPicture(self, updating):
    picture = '\n'.join(self.SimulateDelivery())
    if updating:
      prefix = '\033[F' * len(picture.split('\n'))
    else:
      prefix = ''
    return prefix + picture
 


#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3
import unittest
from typing import Dict, Tuple, List

from sqlalchemy import MetaData, Table, Column, Integer, create_engine

from type_inference.inspectors.postgres_inspector import PostgresInspector
from type_inference.inspectors.sqlite_inspector import SQLiteInspector
from type_inference.inspectors.table_not_exist_exception import TableDoesNotExistException
from type_inference.type_inference_service import TypeInference
from type_inference.types.edge import Equality
from type_inference.types.expression import Variable, PredicateAddressing
from type_inference.types.types_graph import TypesGraph
from type_inference.types.variable_types import NumberType

sqlite_db_file_name = 'logica.db'
number = NumberType()


class TestTypeInferenceWithSqlite(unittest.TestCase):
  @staticmethod
  def create_table(table_name: str, columns: Dict[str, str]):
    with sqlite3.connect(sqlite_db_file_name) as conn:
      columns_to_create = ', '.join([f'{column[0]} {column[1]}' for column in columns.items()])
      cursor = conn.cursor()
      cursor.execute(f'create table if not exists {table_name} (id integer, {columns_to_create})').fetchall()

  @staticmethod
  def safe_drop_table(table_name: str):
    with sqlite3.connect(sqlite_db_file_name) as conn:
      conn.cursor().execute(f'drop table if exists {table_name}').fetchall()

  def test_when_linked_with_predicate_from_db(self):
    # 'Q(x) :- T(x)'
    self.create_table('T', {'col0': 'integer'})
    graph = TypesGraph()
    q_col0 = Variable('col0')
    t_col0 = PredicateAddressing('T', 'col0')
    x_var = Variable('x')
    graph.Connect(Equality(q_col0, x_var, (0, 0)))
    graph.Connect(Equality(x_var, t_col0, (0, 0)))
    graphs = dict()
    graphs['Q'] = graph
    sqlite_inspector = SQLiteInspector('../tests/logica.db')

    TypeInference(graphs, sqlite_inspector).Infer()

    self.assertEquals(q_col0.type, number)
    self.assertEquals(t_col0.type, number)

    self.safe_drop_table('T')

  def test_when_linked_with_unknown_predicate(self):
    # 'Q(x) :- T(x)'
    graph = TypesGraph()
    q_col0 = Variable('col0')
    t_col0 = PredicateAddressing('T', 'col0')
    x_var = Variable('x')
    graph.Connect(Equality(q_col0, x_var, (0, 0)))
    graph.Connect(Equality(x_var, t_col0, (0, 0)))
    graphs = dict()
    graphs['Q'] = graph
    sqlite_inspector = SQLiteInspector('../tests/logica.db')

    with self.assertRaises(TableDoesNotExistException):
      TypeInference(graphs, sqlite_inspector).Infer()


class TestTypeInferenceWithPsql(unittest.TestCase):
  @staticmethod
  def create_table(table_name: str, columns: List[Column]) -> Tuple[MetaData, Table]:
    metadata = MetaData()
    table = Table(table_name, metadata, *columns)
    engine = create_engine(f'postgresql+psycopg2://logica:logica@127.0.0.1', pool_recycle=3600)
    metadata.create_all(engine)
    return metadata, table

  @staticmethod
  def safe_drop_table(metadata: MetaData, table: Table):
    metadata.remove(table)

  def test_when_linked_with_predicate_from_db(self):
    # 'Q(x) :- T(x)'
    metadata, table = self.create_table('T', [Column('col0', Integer, primary_key=True)])
    graph = TypesGraph()
    q_col0 = Variable('col0')
    t_col0 = PredicateAddressing('T', 'col0')
    x_var = Variable('x')
    graph.Connect(Equality(q_col0, x_var, (0, 0)))
    graph.Connect(Equality(x_var, t_col0, (0, 0)))
    graphs = dict()
    graphs['Q'] = graph
    postgres_inspector = PostgresInspector('logica', 'logica')

    TypeInference(graphs, postgres_inspector).Infer()

    self.assertEquals(q_col0.type, number)
    self.assertEquals(t_col0.type, number)

    self.safe_drop_table(metadata, table)

  def test_when_linked_with_unknown_predicate_psql(self):
    # 'Q(x) :- T(x)'
    graph = TypesGraph()
    q_col0 = Variable('col0')
    t_col0 = PredicateAddressing('T', 'col0')
    x_var = Variable('x')
    graph.Connect(Equality(q_col0, x_var, (0, 0)))
    graph.Connect(Equality(x_var, t_col0, (0, 0)))
    graphs = dict()
    graphs['Q'] = graph
    postgres_inspector = PostgresInspector('logica', 'logica')

    with self.assertRaises(TableDoesNotExistException):
      TypeInference(graphs, postgres_inspector).Infer()



#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from type_inference import bq_type_parser

from . import bq_type_parser

if '.' not in __package__:
  from type_inference import type_retriever_base
  from type_inference import unknown_bigquery_type_exception
else:
  from ..type_inference import type_retriever_base
  from ..type_inference import unknown_bigquery_type_exception


class BigQueryTypeRetriever(type_retriever_base.TypeRetrieverBase):
  def __init__(self):
    super().__init__()
    self.array_regexp = re.compile(r'ARRAY<(.*)>')
    self.struct_regexp = re.compile(r'STRUCT<(.*)>')

  def UnpackType(self, type: str) -> str:
    def ParseBigQueryStruct(type):
      nesting_level = 0
      start_index = 0
      fields = {}
      field_name = ''

      for index, char in enumerate(type):
        if char == '<':
          nesting_level += 1
        elif char == '>':
          nesting_level -= 1
        elif nesting_level == 0:
          if char == ' ':
            field_name = type[start_index:index].lstrip()
            start_index = index + 1
          elif char == ',':
            fields[field_name] = type[start_index:index].lstrip()
            start_index = index + 1

      fields[field_name] = type[start_index:].lstrip()
      return fields

    result = bq_type_parser.BigQueryTypeToLogicaType(type)

    if result:
      return result

    array_match = self.array_regexp.match(type)

    if array_match:
      return '[%s]' % self.UnpackTypeWithCaching(array_match.group(1))

    struct_match = self.struct_regexp.match(type)

    if struct_match:
      fields_dict = ParseBigQueryStruct(struct_match.group(1))
      fields = (f'{field_name}: {self.UnpackTypeWithCaching(field_type)}'
                for field_name, field_type in fields_dict.items())
      return '{%s}' % (', '.join(fields))

    raise unknown_bigquery_type_exception.UnknownBigQueryTypeException(type)



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Terminal text color string constants."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

CHR_ERROR = '\033[91m'
CHR_WARNING = '\033[1m'
# TODO: Make dark terminal users be able to configure to use the
# bright yellow.
# CHR_WARNING = '\033[93m'
CHR_UNDERLINE = '\033[4m'
CHR_END = '\033[0m'
CHR_OK = '\033[92m'


def Warn(message):
  return CHR_WARNING + message + CHR_END


def Color(color):
  return ColorsMap()[color]


def ColorsMap():
  return  {'error': CHR_ERROR,
           'warning': CHR_WARNING,
           'underline': CHR_UNDERLINE,
           'ok': CHR_OK,
           'end': CHR_END}


def Format(pattern, args_dict=None):
  args_dict = args_dict or {}
  return pattern.format(**dict(args_dict, **ColorsMap()))



#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from parser_py import parse
from type_inference.types import edge, expression
from type_inference.types_graph_builder import TypesGraphBuilder


class TestTypesGraphBuilding(unittest.TestCase):
  def test_when_connection_with_other_predicates(self):
    s = 'Q(x) :- T(x), Num(x)'
    parsed = parse.ParseFile(s)

    graph = TypesGraphBuilder().Run(parsed)['Q']
    edges = graph.ToEdgesSet()

    expected = [edge.Equality(expression.PredicateAddressing('Q', 'col0'), expression.Variable('x'), (2, 3)),
                edge.Equality(expression.Variable('x'), expression.PredicateAddressing('T', 'col0'), (10, 11)),
                edge.Equality(expression.Variable('x'), expression.PredicateAddressing('Num', 'col0'), (18, 19))]

    self.assertCountEqual(edges, expected)

  def test_when_plus_operator(self):
    s = 'Q(x + y) :- T(x), T(y);'
    parsed = parse.ParseFile(s)

    graph = TypesGraphBuilder().Run(parsed)['Q']
    edges = graph.ToEdgesSet()

    x_var = expression.Variable('x')
    y_var = expression.Variable('y')
    expected = [edge.Equality(expression.PredicateAddressing('Q', 'col0'),
                              expression.PredicateAddressing('+', 'logica_value'), (2, 7)),
                edge.PredicateArgument(expression.PredicateAddressing('+', 'logica_value'),
                                       expression.PredicateAddressing('+', 'left'), (2, 3)),
                edge.PredicateArgument(expression.PredicateAddressing('+', 'logica_value'),
                                       expression.PredicateAddressing('+', 'right'), (6, 7)),
                edge.Equality(x_var, expression.PredicateAddressing('+', 'left'), (2, 3)),
                edge.Equality(y_var, expression.PredicateAddressing('+', 'right'), (6, 7)),
                edge.Equality(x_var, expression.PredicateAddressing('T', 'col0'), (14, 15)),
                edge.Equality(y_var, expression.PredicateAddressing('T', 'col0', 1), (20, 21))]

    self.assertCountEqual(edges, expected)

  def test_when_str(self):
    s = 'Q(x) :- T(x), T(y), Str(x), x == y;'
    parsed = parse.ParseFile(s)

    graph = TypesGraphBuilder().Run(parsed)['Q']
    edges = graph.ToEdgesSet()

    x_var = expression.Variable('x')
    y_var = expression.Variable('y')
    expected = [edge.Equality(expression.PredicateAddressing('Q', 'col0'), expression.Variable('x'), (2, 3)),
                edge.Equality(x_var, expression.PredicateAddressing('T', 'col0'), (10, 11)),
                edge.Equality(y_var, expression.PredicateAddressing('T', 'col0', 1), (16, 17)),
                edge.Equality(x_var, expression.PredicateAddressing('Str', 'col0'), (24, 25)),
                edge.Equality(x_var, y_var, (28, 34))]

    self.assertCountEqual(edges, expected)

  def test_when_concat_operator(self):
    s = 'Q(x ++ y) :- T(x), T(y);'
    parsed = parse.ParseFile(s)

    graph = TypesGraphBuilder().Run(parsed)['Q']
    edges = graph.ToEdgesSet()

    x_var = expression.Variable('x')
    y_var = expression.Variable('y')
    expected = [edge.Equality(expression.PredicateAddressing('Q', 'col0'),
                              expression.PredicateAddressing('++', 'logica_value'), (2, 8)),
                edge.PredicateArgument(expression.PredicateAddressing('++', 'logica_value'),
                                       expression.PredicateAddressing('++', 'left'), (2, 3)),
                edge.PredicateArgument(expression.PredicateAddressing('++', 'logica_value'),
                                       expression.PredicateAddressing('++', 'right'), (7, 8)),
                edge.Equality(x_var, expression.PredicateAddressing('++', 'left'), (2, 3)),
                edge.Equality(y_var, expression.PredicateAddressing('++', 'right'), (7, 8)),
                edge.Equality(x_var, expression.PredicateAddressing('T', 'col0'), (15, 16)),
                edge.Equality(y_var, expression.PredicateAddressing('T', 'col0', 1), (21, 22))]

    self.assertCountEqual(edges, expected)

  def test_when_in_operator(self):
    s = 'Q(y) :- T(x), y in x, Num(y);'
    parsed = parse.ParseFile(s)

    graph = TypesGraphBuilder().Run(parsed)['Q']
    edges = graph.ToEdgesSet()

    x_var = expression.Variable('x')
    y_var = expression.Variable('y')
    expected = [edge.Equality(expression.PredicateAddressing('Q', 'col0'), y_var, (2, 3)),
                edge.Equality(x_var, expression.PredicateAddressing('T', 'col0'), (10, 11)),
                edge.Equality(y_var, expression.PredicateAddressing('Num', 'col0'), (26, 27)),
                edge.EqualityOfElement(x_var, y_var, (14, 20))]

    self.assertCountEqual(edges, expected)

  def test_when_record(self):
    s = 'Q(p: Str(y), q: z + w, s: x) :- T(x), y == x.a, z == x.b, w == x.c.d;'
    parsed = parse.ParseFile(s)

    graph = TypesGraphBuilder().Run(parsed)['Q']
    edges = graph.ToEdgesSet()

    p_var = expression.PredicateAddressing('Q', 'p')
    q_var = expression.PredicateAddressing('Q', 'q')
    s_var = expression.PredicateAddressing('Q', 's')
    y_var = expression.Variable('y')
    z_var = expression.Variable('z')
    w_var = expression.Variable('w')
    x_var = expression.Variable('x')

    expected = [edge.Equality(p_var, expression.PredicateAddressing('Str', 'logica_value'), (5, 10)),
                edge.PredicateArgument(expression.PredicateAddressing('Str', 'logica_value'),
                                       expression.PredicateAddressing('Str', 'col0'), (9, 10)),
                edge.Equality(y_var, expression.PredicateAddressing('Str', 'col0'), (9, 10)),
                edge.Equality(q_var, expression.PredicateAddressing('+', 'logica_value'), (16, 21)),
                edge.PredicateArgument(expression.PredicateAddressing('+', 'logica_value'),
                                       expression.PredicateAddressing('+', 'left'), (16, 17)),
                edge.PredicateArgument(expression.PredicateAddressing('+', 'logica_value'),
                                       expression.PredicateAddressing('+', 'right'), (20, 21)),
                edge.Equality(z_var, expression.PredicateAddressing('+', 'left'), (16, 17)),
                edge.Equality(w_var, expression.PredicateAddressing('+', 'right'), (20, 21)),
                edge.Equality(s_var, x_var, (26, 27)),
                edge.Equality(x_var, expression.PredicateAddressing('T', 'col0'), (34, 35)),
                edge.Equality(expression.SubscriptAddressing(x_var, 'a'), y_var, (38, 46)),
                edge.Equality(expression.SubscriptAddressing(x_var, 'b'), z_var, (48, 56)),
                edge.Equality(expression.SubscriptAddressing(expression.SubscriptAddressing(x_var, 'c'), 'd'), w_var,
                              (58, 68)),
                edge.FieldBelonging(x_var, expression.SubscriptAddressing(x_var, 'a'), (43, 46)),
                edge.FieldBelonging(x_var, expression.SubscriptAddressing(x_var, 'b'), (53, 56)),
                edge.FieldBelonging(x_var, expression.SubscriptAddressing(x_var, 'c'), (63, 66)),
                edge.FieldBelonging(expression.SubscriptAddressing(x_var, 'c'),
                                    expression.SubscriptAddressing(expression.SubscriptAddressing(x_var, 'c'), 'd'),
                                    (63, 68))]

    self.assertCountEqual(edges, expected)

  def test_when_named_columns(self):
    s = 'Q(a:, b:) :- T(x), T(y), a == x + y, b == x + y;'
    parsed = parse.ParseFile(s)

    graph = TypesGraphBuilder().Run(parsed)['Q']
    edges = graph.ToEdgesSet()

    x_var = expression.Variable('x')
    y_var = expression.Variable('y')
    a_var = expression.Variable('a')
    b_var = expression.Variable('b')
    expected = [edge.Equality(expression.PredicateAddressing('Q', 'a'), a_var, (2, 3)),
                edge.Equality(expression.PredicateAddressing('Q', 'b'), b_var, (6, 7)),
                edge.Equality(x_var, expression.PredicateAddressing('T', 'col0'), (15, 16)),
                edge.Equality(y_var, expression.PredicateAddressing('T', 'col0', 1), (21, 22)),
                edge.Equality(a_var, expression.PredicateAddressing('+', 'logica_value'), (25, 35)),
                edge.Equality(b_var, expression.PredicateAddressing('+', 'logica_value', 1), (37, 47)),
                edge.PredicateArgument(expression.PredicateAddressing('+', 'logica_value'),
                                       expression.PredicateAddressing('+', 'left'), (30, 31)),
                edge.PredicateArgument(expression.PredicateAddressing('+', 'logica_value'),
                                       expression.PredicateAddressing('+', 'right'), (34, 35)),
                edge.PredicateArgument(expression.PredicateAddressing('+', 'logica_value', 1),
                                       expression.PredicateAddressing('+', 'left', 1), (42, 43)),
                edge.PredicateArgument(expression.PredicateAddressing('+', 'logica_value', 1),
                                       expression.PredicateAddressing('+', 'right', 1), (46, 47)),
                edge.Equality(x_var, expression.PredicateAddressing('+', 'left'), (30, 31)),
                edge.Equality(y_var, expression.PredicateAddressing('+', 'right'), (34, 35)),
                edge.Equality(x_var, expression.PredicateAddressing('+', 'left', 1), (42, 43)),
                edge.Equality(y_var, expression.PredicateAddressing('+', 'right', 1), (46, 47))]

    self.assertCountEqual(edges, expected)



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unitests for platonic.py."""

from common.platonic import *
import unittest


class PlatonicTest(unittest.TestCase):
  def test_AdherenceSuccess(self):
    t1 = PlatonicObject({'a': Num})
    self.assertEqual(
        t1.AdheredBy({'a': 1}).Elements(), ({'a': 1}, True, None))

    t2 = PlatonicObject(
        {'a': Num,
         'aa': PlatonicObject({'b': Str,
                              'c': PlatonicObject({'d': Str})})})
    x2 = {'a': 1.0, 'aa': {'b': 'Seattle', 'c': {'d': 'Washington'}}}
    self.assertEqual(
        t2.AdheredBy(x2).Elements(), (x2, True, None))

    t3 = PlatonicObject(
        {'a': [Num], 'b': PlatonicObject({'c': [PlatonicObject({'d': Num})]})}
    )
    x3 = {'a': [1], 'b': {'c': [{'d': 5}]}}
    self.assertEqual(
        t3.AdheredBy(x3).Elements(), (x3, True, None))

  def test_AdherenceFailure(self):
    t1 = PlatonicObject({'a': Str})
    x1 = {'a': 1}
    self.assertEqual(
        t1.AdheredBy(x1).Elements(),
        (1, False, "{'a': 1}.a: Expected str, got 1"))

    t2 = PlatonicObject(
        {'a': Num,
         'aa': PlatonicObject({'b': Str,
                               'c': PlatonicObject({'d': Str})})})
    x2 = {'a': 1.0, 'aa': {'b': 5, 'c': {'d': 'Washington'}}}
    self.assertEqual(
        t2.AdheredBy(x2).Elements(),
        (5, False,
         "{'a': 1.0, 'aa': {'b': 5, 'c': {'d': " +
         "'Washington'}}}.aa.b: Expected str, got 5"))

    t3 = PlatonicObject(
        {'a': [Num], 'b': PlatonicObject({'c': [PlatonicObject({'d': Num})]})}
    )
    x3 = {'a': [1], 'b': {'c': {'d': 5}}}
    self.assertEqual(
        t3.AdheredBy(x3).Elements(),
        ({'c': {'d': 5}}, False, 
         "Record {'a': [1], 'b': {'c': {'d': 5}}}, " +
         "at path: .b encountered error: " +
         "Field value c was expected to be a list, got: {'d': 5}."))

    t4 = PlatonicObject({'a': PlatonicObject({'b': 1})})
    x4 = {'a': 5}
    self.assertEqual(t4.AdheredBy(x4).Elements(),
                     (5, False, 
                      "Record {'a': 5}, at path: .a encountered error:" +
                      " Expected record, but got not a dictionary."))
  
  def test_CheckPasses(self):
    t1 = PlatonicObject({'a': Str})
    x1 = {'a': 'earth'}
    self.assertEqual(t1(x1), x1)

  def test_CheckFails(self):
    t1 = PlatonicObject({'a': Str})
    x1 = {'a': 42}
    with self.assertRaises(AssertionError):
      t1(x1)


#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List

from type_inference.built_in_functions_types import built_in_restrictions
from type_inference.types.variable_types import AnyType, NumberType, StringType, RecordType, ListType


class Expression:
  def __init__(self):
    self.type = AnyType()
    self.__class__.__hash__ = Expression.__hash__

  def __eq__(self, other):
    return isinstance(other, type(self))

  def __hash__(self):
    return hash(str(self))

  def __str__(self):
    return type(self).__name__


class PredicateAddressing(Expression):
  def __init__(self, predicate_name: str, field: str, predicate_id: int = 0):
    super().__init__()
    self.predicate_name = predicate_name
    self.field = field
    self.predicate_id = predicate_id

    if predicate_name in built_in_restrictions:
      self.type = built_in_restrictions[predicate_name][self.field]

  def __eq__(self, other):
    return super().__eq__(other) and \
           (self.predicate_name, self.field, self.predicate_id) == \
           (other.predicate_name, other.field, other.predicate_id)

  def __str__(self):
    return super().__str__() + f'({self.predicate_name}.{self.field})'


class SubscriptAddressing(Expression):
  def __init__(self, base: Expression, subscript_field: str):
    super().__init__()
    self.base = base
    self.subscript_field = subscript_field

  def __eq__(self, other):
    return super().__eq__(other) and self.subscript_field == other.subscript_field

  def __str__(self):
    return super().__str__() + f'{str(self.base)}.{self.subscript_field}'


class Variable(Expression):
  def __init__(self, variable_name):
    super().__init__()
    self.variable_name = variable_name

  def __eq__(self, other):
    return super().__eq__(other) and self.variable_name == other.variable_name

  def __str__(self):
    return super().__str__() + f'({self.variable_name})'


class Literal(Expression):
  pass


class StringLiteral(Literal):
  def __init__(self):
    super().__init__()
    self.type = StringType()


class NumberLiteral(Literal):
  def __init__(self):
    super().__init__()
    self.type = NumberType()


class BooleanLiteral(Literal):
  pass


class ListLiteral(Literal):
  def __init__(self, elements: List[Expression]):
    super().__init__()
    self.elements = elements
    if len(elements) > 0:
      self.type = ListType(elements[0].type)
    else:
      self.type = ListType(AnyType())

  def __eq__(self, other):
    return super().__eq__(other) and self.elements == self.elements

  def __str__(self):
    return super().__str__() + f'[{", ".join(map(str, self.elements))}]'


class NullLiteral(Literal):
  pass


class RecordLiteral(Literal):
  def __init__(self, fields: Dict[str, Expression]):
    super().__init__()
    self.fields = fields
    self.type = RecordType({name: expr.type for name, expr in fields.items()}, False)

  def __eq__(self, other):
    return super().__eq__(other) and self.fields == other.fields

  def __str__(self):
    fields = ', '.join(map(lambda t: f'{t[0]}: {t[1]}', sorted(self.fields.items(), key=lambda t: t[0])))
    return super().__str__() + f'({fields})'



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility for drawing images."""

import numpy
from matplotlib import pyplot
from matplotlib import animation

import IPython
from PIL import Image
import glob


def Animate(spacetime,
            column_x='col0',
            column_y='col1',
            column_t='col2',
            column_v='logica_value',
            field_width=None,
            field_height=None,
            cmap='hot',
            vmin=0,
            vmax=None,
            shadow_factor=0.5,
            min_shadow=0.1,
            gif_file='animation.gif',
            should_display=True,
            fps=10,
            figsize=None,
            show_progress=True):
  """Animating a dataframe."""
  fig, ax = pyplot.subplots(figsize=figsize)
  if field_width is None:
    field_width = max(spacetime[column_x]) + 1
  if field_height is None:
    field_height = max(spacetime[column_y]) + 1  
  field_shape = (field_width, field_height)

  if column_v is None:
    column_v = '*magical_indicator_column*'
    spacetime[column_v] = 1
  if column_v == 'logica_value' and 'logica_value' not in spacetime:
    spacetime['logica_value'] = 1

  if vmin is None:
    vmin = min(spacetime[column_v])
  if vmax is None:
    vmax = max(spacetime[column_v])
  r = numpy.zeros(field_shape)
  img = pyplot.imshow(r.T, cmap=cmap, vmin=vmin, vmax=vmax)

  num_frames = max(spacetime[column_t])
  def Update(frame):
    if show_progress:
      total = 40
      done = frame * 40 // (num_frames - 1)
      print('\rAnimating: ' + done * '▒' + (total - done) * '░',
            end='', flush=True)
    r = numpy.zeros(field_shape)
    decay = 1
    i = 0
    while decay > min_shadow:
      m = numpy.zeros(field_shape)
      df = spacetime[spacetime[column_t] == (frame - i)]
      m[df[column_x], df[column_y]] = df[column_v] * decay
      r = numpy.maximum(r, m)
      decay *= shadow_factor
      i += 1
    img.set_data(r.T)
  ani = animation.FuncAnimation(fig, Update, frames=range(num_frames))
  if show_progress:
    print('')

  ani.save(gif_file, writer='pillow', fps=fps)
  if should_display:
    IPython.display.display(IPython.display.Image(filename=gif_file))
  pyplot.close(fig)

def ShowImage(img,
              column_x='col0',
              column_y='col1',
              column_v='logica_value',
              field_width=None,
              field_height=None,
              cmap='hot',
              vmin=0,
              vmax=None):
  """Showing dataframe as an image."""
  if column_v is None or (column_v == 'logica_value' and
                          column_v not in img):
    if column_v is None:
      column_v = '*magical_indicator_column*'
    img[column_v] = numpy.zeros(len(img)) + 1
  if field_width is None:
    field_width = max(img[column_x]) + 1
  if field_height is None:
    field_height = max(img[column_y]) + 1  
  field_shape = (field_width, field_height)

  if vmin is None:
    vmin = min(img[column_v])
  if vmax is None:
    vmax = max(img[column_v])
  r = numpy.zeros(field_shape)
  r[img[column_x], img[column_y]] = img[column_v]
  pyplot.imshow(r.T, cmap=cmap, vmin=vmin, vmax=vmax)

def Hexagonal(img,
              column_x='col0',
              column_y='col1',
              column_v='logica_value',
              field_width=None,
              field_height=None,
              cmap='hot',
              vmin=0,
              vmax=None,
              figsize=None):
  """Showing dataframe as an image."""
  if column_v is None or (column_v == 'logica_value' and
                          column_v not in img):
    if column_v is None:
      column_v = '*magical_indicator_column*'
    img[column_v] = numpy.zeros(len(img)) + 1
  if field_width is None:
    field_width = max(img[column_x]) + 1
  if field_height is None:
    field_height = max(img[column_y]) + 1  
  field_shape = (field_width, field_height)

  if vmin is None:
    vmin = min(img[column_v])
  if vmax is None:
    vmax = max(img[column_v])

  hex_values = numpy.zeros(field_shape)
  hex_values[img[column_x], img[column_y]] = (
      img[column_v] - vmin) / (vmax - vmin)

  cmap = pyplot.get_cmap('hot')
  colors = cmap(hex_values)

  # Computing hexagon.
  π = 3.14159265
  angles = numpy.arange(0, 2 * π, π / 3)
  hex_delta = numpy.vstack([numpy.cos(angles), numpy.sin(angles)]).T

  def DrawOneHexagon(center, ax, color):
    hexagon = center + hex_delta
    ax.fill(hexagon[:,0], hexagon[:,1], c=color, edgecolor=color)

  fig, ax = pyplot.subplots(figsize=figsize, facecolor='black')
  ax.set_facecolor('black')

  # Get coordinates of centers.
  rows, cols = numpy.indices((field_width, field_height))
  rows = rows.flatten()
  cols = cols.flatten()
  xs = 1.5 * rows
  ys = 3 ** 0.5 * (cols + 0.5 * (rows % 2))
  centers = numpy.vstack([xs,
                          ys]).T
  
  # Draw.
  for center, color in zip(centers, colors.reshape(-1, colors.shape[-1])):
    DrawOneHexagon(center, ax, color)
  
  # Prettify.
  ax.set_aspect('equal')
  pyplot.xticks([])
  pyplot.yticks([])

def AnimateFiles(images_glob,
                 output_file,
                 duration=100,
                 and_back=False):
  imgs = [Image.open(f) for f in sorted(glob.glob(images_glob))]
  if and_back:
    imgs += reversed(imgs)
  imgs[0].save(fp=output_file, format='GIF', append_images=imgs[1:],
               save_all=True, duration=duration, loop=0)


#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Dict, Tuple, Union

from parser_py.parse import HeritageAwareString
from type_inference.types.edge import Equality, EqualityOfElement, FieldBelonging, PredicateArgument
from type_inference.types.expression import StringLiteral, NumberLiteral, BooleanLiteral, NullLiteral, ListLiteral, \
  PredicateAddressing, SubscriptAddressing, Variable, RecordLiteral, Literal, Expression
from type_inference.types.types_graph import TypesGraph


class TypesGraphBuilder:
  _predicate_usages: defaultdict
  _if_statements_counter: int
  _expressions_cache: dict

  def ResetInternalState(self):
    self._predicate_usages = defaultdict(lambda: 0)
    self._if_statements_counter = 0
    self._expressions_cache = {}

  def GetFromCacheOrAdd(self, expression: Expression):
    if expression not in self._expressions_cache:
      self._expressions_cache[expression] = expression

    return self._expressions_cache[expression]

  def Run(self, parsed_program: dict) -> Dict[str, TypesGraph]:
    self.ResetInternalState()
    graphs = defaultdict(lambda: TypesGraph())

    for rule in parsed_program['rule']:
      predicate_name = rule['head']['predicate_name']
      graphs[predicate_name] |= self.TraverseTree(predicate_name, rule)

    return graphs

  def TraverseTree(self, predicate_name: str, rule: dict) -> TypesGraph:
    types_graph = TypesGraph()

    for field in rule['head']['record']['field_value']:
      self.FillField(types_graph, predicate_name, field)

    if 'body' in rule:
      for conjunct in rule['body']['conjunction']['conjunct']:
        self.FillConjunct(types_graph, conjunct)

    return types_graph

  def FillField(self, types_graph: TypesGraph, predicate_name: str, field: dict):
    field_name = field['field']

    if isinstance(field_name, int):
      field_name = f'col{field_name}'

    variable = self.GetFromCacheOrAdd(
      PredicateAddressing(predicate_name, field_name, self._predicate_usages[predicate_name]))

    if 'aggregation' in field['value']:
      value, bounds = self.ConvertExpression(types_graph, field['value']['aggregation']['expression'])
      types_graph.Connect(Equality(variable, value, bounds))
      return

    if 'expression' in field['value']:
      value, bounds = self.ConvertExpression(types_graph, field['value']['expression'])
      types_graph.Connect(Equality(variable, value, bounds))
      return

    raise NotImplementedError(field)

  def FillConjunct(self, types_graph: TypesGraph, conjunct: dict):
    if 'unification' in conjunct:
      unification = conjunct['unification']
      left_hand_side, (left, _) = self.ConvertExpression(types_graph, unification['left_hand_side'])
      right_hand_side, (_, right) = self.ConvertExpression(types_graph, unification['right_hand_side'])
      types_graph.Connect(Equality(left_hand_side, right_hand_side, (left, right)))
    elif 'inclusion' in conjunct:
      inclusion = conjunct['inclusion']
      list_of_elements, (_, right) = self.ConvertExpression(types_graph, inclusion['list'])
      element, (left, _) = self.ConvertExpression(types_graph, inclusion['element'])
      types_graph.Connect(EqualityOfElement(list_of_elements, element, (left, right)))
    elif 'predicate' in conjunct:
      value = conjunct['predicate']
      predicate_name = value['predicate_name']
      logica_value = PredicateAddressing(predicate_name, 'logica_value', self._predicate_usages[predicate_name])
      self.FillFields(predicate_name, types_graph, value, logica_value)
      self._predicate_usages[predicate_name] += 1
    else:
      raise NotImplementedError(conjunct)

  def FillFields(self, predicate_name: str, types_graph: TypesGraph, fields: dict, result: PredicateAddressing) -> \
          Tuple[int, int]:
    total_min = None
    total_max = None

    for field in fields['record']['field_value']:
      value, bounds = self.ConvertExpression(types_graph, field['value']['expression'])
      total_min = MinIgnoringNone(total_min, bounds[0])
      total_max = MaxIgnoringNone(total_max, bounds[1])
      field_name = field['field']

      if isinstance(field_name, int):
        field_name = f'col{field_name}'

      predicate_field = PredicateAddressing(predicate_name, field_name, self._predicate_usages[predicate_name])
      types_graph.Connect(Equality(predicate_field, value, bounds))
      types_graph.Connect(PredicateArgument(result, predicate_field, bounds))

    return total_min, total_max

  def ConvertExpression(self, types_graph: TypesGraph, expression: dict) -> Tuple[Expression, Tuple[int, int]]:
    if 'literal' in expression:
      result, bounds = self.ConvertLiteralExpression(types_graph, expression['literal'])
    elif 'variable' in expression:
      value = expression['variable']['var_name']
      result, bounds = Variable(value), (value.start, value.stop)
    elif 'call' in expression:
      result, bounds = self.ConvertCallExpression(types_graph, expression['call'])
    elif 'subscript' in expression:
      result, bounds = self.ConvertSubscriptExpression(types_graph, expression['subscript'])
    elif 'record' in expression:
      result, bounds = self.ConvertRecordExpression(types_graph, expression['record']['field_value'])
    elif 'implication' in expression:
      result, bounds = self.ConvertImplicationExpression(types_graph, expression['implication'])
    else:
      raise NotImplementedError(expression)

    return self.GetFromCacheOrAdd(result), bounds

  def ConvertLiteralExpression(self, types_graph: TypesGraph, literal: dict) -> Tuple[Literal, Tuple[int, int]]:
    if 'the_string' in literal:
      string = literal['the_string']['the_string']
      return StringLiteral(), (string.start, string.stop)
    elif 'the_number' in literal:
      number = literal['the_number']['number']
      return NumberLiteral(), (number.start, number.stop)
    elif 'the_bool' in literal:
      boolean = literal['the_bool']['bool']
      return BooleanLiteral(), (boolean.start, boolean.stop)
    elif 'the_null' in literal:
      null = literal['the_null']['null']
      return NullLiteral(), (null.start, null.stop)
    elif 'the_list' in literal:
      total_min = None
      total_max = None
      elements = []

      for element in literal['the_list']['element']:
        expression, (left, right) = self.ConvertExpression(types_graph, element)
        elements.append(expression)
        total_min = MinIgnoringNone(total_min, left)
        total_max = MinIgnoringNone(total_max, right)

      return ListLiteral(elements), (total_min - 1, total_max + 1)

  def ConvertCallExpression(self, types_graph: TypesGraph, call: dict):
    predicate_name = call['predicate_name']
    result = self.GetFromCacheOrAdd(
      PredicateAddressing(predicate_name, 'logica_value', self._predicate_usages[predicate_name]))
    bounds = self.FillFields(predicate_name, types_graph, call, result)
    self._predicate_usages[predicate_name] += 1
    return result, self.MoveBoundsAccordingToPredicateNameType(bounds, predicate_name)

  def ConvertSubscriptExpression(self, types_graph: TypesGraph, subscript: dict):
    record, (left, _) = self.ConvertExpression(types_graph, subscript['record'])
    field = subscript['subscript']['literal']['the_symbol']['symbol']
    result = self.GetFromCacheOrAdd(SubscriptAddressing(record, field))
    bounds = (left, field.stop)
    types_graph.Connect(FieldBelonging(record, result, bounds))
    return result, bounds

  def ConvertRecordExpression(self, types_graph: TypesGraph, field_value: dict):
    fields = {}
    total_min, total_max = None, None

    for field in field_value:
      expression, (left, right) = self.ConvertExpression(types_graph, field['value']['expression'])
      fields[field['field']] = expression
      total_min = MinIgnoringNone(total_min, left)
      total_max = MaxIgnoringNone(total_max, right)

    return RecordLiteral(fields), (total_min - 1, total_max + 1)

  def ConvertImplicationExpression(self, types_graph: TypesGraph, implication: dict):
    inner_variable = Variable(f'_IfNode{self._if_statements_counter}')
    self._if_statements_counter += 1
    otherwise, (common_left, common_right) = self.ConvertExpression(types_graph, implication['otherwise'])
    types_graph.Connect(Equality(inner_variable, otherwise, (common_left, common_right)))

    for i in implication['if_then']:
      self.ConvertExpression(types_graph, i['condition'])
      value, (left, right) = self.ConvertExpression(types_graph, i['consequence'])
      types_graph.Connect(Equality(inner_variable, value, (left, right)))
      common_left = MinIgnoringNone(common_left, left)
      common_right = MinIgnoringNone(common_right, right)

    return inner_variable, (common_left, common_right)

  def MoveBoundsAccordingToPredicateNameType(self, bounds: Tuple[int, int],
                                             predicate_name: Union[str, HeritageAwareString]) -> Tuple[int, int]:
    if isinstance(predicate_name, HeritageAwareString):
      return MinIgnoringNone(bounds[0], predicate_name.start), MaxIgnoringNone(bounds[1], predicate_name.stop)

    return bounds


def MinIgnoringNone(left: Union[int, None], right: int) -> int:
  return min(left, right) if left else right


def MaxIgnoringNone(left: Union[int, None], right: int) -> int:
  return max(left, right) if left else right



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Example of running Logica on PostgreSQL from Python.
# Additional libraries to be installed if using logica as a library:
# `pip install psycopg2`
# Assuming a PostgreSQL connection shown as a  terminal command below:
#  PGPASSWORD=somepassword psql -d postgres -U postgres -h localhost
# To run: python3 call_postgres.py

from logica.common import logica_lib
from sqlalchemy import create_engine
print('Calculating primes.')
engine = create_engine('postgresql+psycopg2://postgres:somepassword@localhost', pool_recycle=3600)
connection = engine.connect()
primes = logica_lib.RunPredicateToPandas(
    '../scripts/primes_postgresql.l',
    'Prime',
    connection=connection)
print('Found %d primes.' % len(primes))
print('Primes:', list(primes['prime']))






#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

from type_inference.types.expression import Expression, SubscriptAddressing, PredicateAddressing


class Edge:
  def __init__(self, vertices: Tuple[Expression, Expression], bounds: Tuple[int, int]):
    self.vertices = vertices
    self.bounds = bounds

  def __eq__(self, other):
    return isinstance(other, type(self)) and (set(self.vertices), self.bounds) == (set(other.vertices), other.bounds)

  def __hash__(self):
    return hash((frozenset(self.vertices), self.bounds))


class Equality(Edge):
  def __init__(self, left: Expression, right: Expression, bounds: Tuple[int, int]):
    super().__init__((left, right), bounds)
    self.left = left
    self.right = right


class EqualityOfElement(Edge):
  def __init__(self, list: Expression, element: Expression, bounds: Tuple[int, int]):
    super().__init__((list, element), bounds)
    self.list = list
    self.element = element


class FieldBelonging(Edge):
  def __init__(self, parent: Expression, field: SubscriptAddressing, bounds: Tuple[int, int]):
    super().__init__((parent, field), bounds)
    self.parent = parent
    self.field = field


class PredicateArgument(Edge):
  def __init__(self, logica_value: PredicateAddressing, argument: PredicateAddressing, bounds: Tuple[int, int]):
    super().__init__((logica_value, argument), bounds)
    self.logica_value = logica_value
    self.argument = argument



#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from parser_py import parse
from type_inference.types.expression import Variable, PredicateAddressing, SubscriptAddressing
from type_inference.types_graph_builder import TypesGraphBuilder
from type_inference.type_inference_service import TypeInference
from type_inference.types.variable_types import NumberType, StringType, ListType, RecordType, AnyType

class IntegrationTypeInferenceTest(unittest.TestCase):
  def find_val(self, val, graph, check):
    for op in graph.ToEdgesSet():
      for v in op.vertices:
        if check(v, val):
          self.assertEqual(val.type, v.type)
          return
    raise Exception("Value not found")

  @staticmethod
  def check_var(v1, v2):
    return v1 == v2

  @staticmethod
  def check_predicate_addressing(p1, p2):
    return isinstance(p1, PredicateAddressing) and p1.predicate_name == p2.predicate_name and p1.field == p2.field

  @staticmethod
  def check_subscript_addressing(p1, p2):
    return isinstance(p1, SubscriptAddressing) and p1.subscript_field == p2.subscript_field and p1.base == p2.base

  def test_built_in_Num(self):
    raw_program = 'Q(x) :- Num(x)'
    parsed = parse.ParseFile(raw_program)
    graphs = dict()
    graphs['Q'] = TypesGraphBuilder().Run(parsed)['Q']
    TypeInference(graphs).Infer()
    var = Variable('x')
    var.type = NumberType()
    self.find_val(var, graphs['Q'], self.check_var)

  def test_built_in_range(self):
    raw_program = 'Q(x) :- x in Range(10)'
    parsed = parse.ParseFile(raw_program)
    graphs = dict()
    graphs['Q'] = TypesGraphBuilder().Run(parsed)['Q']
    TypeInference(graphs).Infer()
    var = Variable('x')
    var.type = NumberType()
    self.find_val(var, graphs['Q'], self.check_var)

  def test_unknown_type(self):
    raw_program = 'Q(x, y) :- x == y'
    parsed = parse.ParseFile(raw_program)
    graphs = dict()
    graphs['Q'] = TypesGraphBuilder().Run(parsed)['Q']
    TypeInference(graphs).Infer()
    var = Variable('x')
    var.type = AnyType()
    self.find_val(var, graphs['Q'], self.check_var)

  def test_three_predicates(self):
    raw = 'Q(x, y) :- T(y), Num(x), F(x); \n T(x) :- x == 10; \n F(x) :- x == 1; \n'
    parsed = parse.ParseFile(raw)
    res = TypesGraphBuilder().Run(parsed)
    TypeInference(res).Infer()
    var_x = Variable('x')
    var_y = Variable('y')
    var_x.type = NumberType()
    var_y.type = NumberType()
    self.find_val(var_x, res['Q'], self.check_var)
    self.find_val(var_x, res['T'], self.check_var)
    self.find_val(var_x, res['F'], self.check_var)
    self.find_val(var_y, res['Q'], self.check_var)

  def test_one_of_arguments_any(self):
    raw = 'Q(x, y) :- T(x), Num(x), F(x); \n T(x) :- x == 10; \n F(x) :- x == 1; \n'
    parsed = parse.ParseFile(raw)
    res = TypesGraphBuilder().Run(parsed)
    TypeInference(res).Infer()
    var_x = Variable('x')
    var_y = Variable('y')
    var_x.type = NumberType()
    var_y.type = AnyType()
    self.find_val(var_x, res['Q'], self.check_var)
    self.find_val(var_x, res['T'], self.check_var)
    self.find_val(var_x, res['F'], self.check_var)
    self.find_val(var_y, res['Q'], self.check_var)

  def test_when_list(self):
    raw = 'Q(Range(10));'
    parsed = parse.ParseFile(raw)
    res = TypesGraphBuilder().Run(parsed)
    TypeInference(res).Infer()
    predicate = PredicateAddressing('Q', 'col0')
    predicate.type = ListType(NumberType())
    self.find_val(predicate, res['Q'], self.check_predicate_addressing)

  def test_when_return_list(self):
    raw = 'F(x) :- x == Q(); \n Q() = Range(10);'
    parsed = parse.ParseFile(raw)
    res = TypesGraphBuilder().Run(parsed)
    TypeInference(res).Infer()
    var = Variable('x')
    var.type = ListType(NumberType())
    self.find_val(var, res['F'], self.check_var)

  def test_when_opened_record(self):
    raw = 'Q(x) :- x.a == 1, x.b == "string"'
    parsed = parse.ParseFile(raw)
    res = TypesGraphBuilder().Run(parsed)
    TypeInference(res).Infer()
    var = Variable('x')
    var.type = RecordType({'a': NumberType(), 'b': StringType()}, True)
    self.find_val(var, res['Q'], self.check_var)

  def test_when_opened_record_inside_open_record(self):
    raw = 'Q(x) :- x.a == 1, x.b.c == "string", x.b.a == 1;'
    parsed = parse.ParseFile(raw)
    res = TypesGraphBuilder().Run(parsed)
    TypeInference(res).Infer()
    var = Variable('x')
    var.type = RecordType({'a': NumberType(), 'b': RecordType({'c': StringType(), 'a': NumberType()}, True)}, True)
    self.find_val(var, res['Q'], self.check_var)

  def test_when_closed_record(self):
    raw = 'Q({a: 1, b: "string", c: {a :"str", b: 1}})'
    parsed = parse.ParseFile(raw)
    res = TypesGraphBuilder().Run(parsed)
    TypeInference(res).Infer()
    predicate = PredicateAddressing('Q', 'col0')
    predicate.type = RecordType({
      'a': NumberType(),
      'b': StringType(),
      'c': RecordType({'a': StringType(), 'b': NumberType()}, False)},
      False)
    self.find_val(predicate, res['Q'], self.check_predicate_addressing)

  def test_when_value_is_function_result(self):
    raw = 'Q(x) :- x.a == T(), x.b == "string"; \n T() = 777;'
    parsed = parse.ParseFile(raw)
    res = TypesGraphBuilder().Run(parsed)
    TypeInference(res).Infer()
    var = Variable('x')
    var.type = RecordType({'a': NumberType(), 'b': StringType()}, True)
    self.find_val(var, res['Q'], self.check_var)

  def test_when_list_in_closed_record(self):
    raw = 'F(x) :- x == {b: 1, a: [1,2]};'
    parsed = parse.ParseFile(raw)
    res = TypesGraphBuilder().Run(parsed)
    TypeInference(res).Infer()
    var = Variable('x')
    var.type = RecordType({'a': ListType(NumberType()), 'b': NumberType()}, False)
    self.find_val(var, res['F'], self.check_var)



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logica command-line tool.

Example usage:

python3 logica.py - run Grandparent <<<'
Parent(parent: "Shmi Skywalker", child: "Anakin Skywalker");
Parent(parent: "Anakin Skywalker", child: "Luke Skywalker");
Grandparent(grandparent:, grandchild:) :-
  Parent(parent: grandparent, child: x),
  Parent(parent: x, child: grandchild);
'

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import getopt
import json
import os
import subprocess
import sys

# We are doing this 'if' to allow usage of the code as package and as a
# script.
if __name__ == '__main__' and not __package__:
  from common import color
  from common import sqlite3_logica
  from compiler import functors
  from compiler import rule_translate
  from compiler import universe
  from parser_py import parse
  from type_inference.research import infer
  from type_inference import type_retrieval_service_discovery
else:
  from .common import color
  from .common import sqlite3_logica
  from .compiler import functors
  from .compiler import rule_translate
  from .compiler import universe
  from .parser_py import parse
  from .type_inference.research import infer
  from .type_inference import type_retrieval_service_discovery


def ReadUserFlags(rules, argv):
  """Reading logic program flags provided by the user."""
  def Error(msg):
    print(color.Format('[ {error}Error{end} ] {msg}', {'msg': msg}))
    sys.exit(1)

  annotations = universe.Annotations.ExtractAnnotations(
      rules, restrict_to=['@DefineFlag'])
  defined_flags = annotations['@DefineFlag'].keys()
  try:
    p = getopt.getopt(argv, '', ['%s=' % f for f in defined_flags])
  except getopt.GetoptError as e:
    Error(str(e))

  if p[1]:
    Error('Undefined command arguments: %s' % p[1])

    sys.exit(1)
  user_flags = {k[2:]: v for k, v in p[0]}
  return user_flags


def GetImportRoot():
  """Parses LOGICAPATH environment variable."""
  import_root_env = os.environ.get('LOGICAPATH')
  if not import_root_env:
    return None
  roots = import_root_env.split(':')
  if len(roots) > 1:
    return roots
  else:
    return import_root_env


def GetTrinoParameters(a):
  boolean_parameters = [
    "debug", "disable-compression", "ignore-errors",
    "insecure", "krb5-disable-remote-service-hostname-canonicalization",
    "password", "progress"]
  parameters = [
    "access-token", "catalog", "client-info", "client-request-timeout",
    "client-tags", "execute", "external-authentication",
    "extra-credential", "http-proxy", "keystore-password",
    "keystore-path", "keystore-type", "krb5-config-path",
    "krb5-credential-cache-path", "krb5-keytab-path",
    "krb5-principal", "krb5-remote-service-name",
    "krb5-service-principal-pattern", "log-levels-file",
    "resource-estimate", "schema", "server", "session",
    "session-user", "socks-proxy", "source", "timezone", "trace-token",
    "truststore-password", "truststore-path", "truststore-type", "user"]
  boolean_params = ["--%s" % p for p in boolean_parameters
    if (p in a and type(a.get(p)) == bool and a.get(p))]
  params = ["--%s=%s" % (p, a.get(p)) for p in parameters if p in a]
  if "catalog" not in a:
    params.append("--catalog=memory")
  return boolean_params + params


def main(argv):
  if len(argv) <= 1 or argv[1] == 'help':
    print('Usage:')
    print('  logica <l file> <command> <predicate name> [flags]')
    print('  Commands are:')
    print('    print: prints the StandardSQL query for the predicate.')
    print('    run: runs the StandardSQL query on BigQuery with pretty output.')
    print('    run_to_csv: runs the query on BigQuery with csv output.')

    print('')
    print('')
    print('Example:')
    print('  python3 logica.py - run GoodIdea <<<\' '
          'GoodIdea(snack: "carrots")\'')
    return 1

  if len(argv) == 3 and argv[2] in ['parse', 'infer_types', 'show_signatures']:
    pass  # compile needs just 2 actual arguments.
  else:
    if len(argv) < 4:
      print('Not enough arguments. Run \'logica help\' for help.',
            file=sys.stderr)
      return 1
    predicates = argv[3]

  if argv[1] == '-':
    filename = '/dev/stdin'
  else:
    filename = argv[1]

  command = argv[2]

  commands = ['parse', 'print', 'run', 'run_to_csv', 'run_in_terminal',
              'infer_types', 'show_signatures', 'build_schema']

  if command not in commands:
    print(color.Format('Unknown command {warning}{command}{end}. '
                       'Available commands: {commands}.',
                       dict(command=command, commands=', '.join(commands))))
    return 1
  if not os.path.exists(filename):
    print('File not found: %s' % filename, file=sys.stderr)
    return 1

  # This has to be before reading program.
  if command == 'run_in_terminal':
    if __name__ == '__main__' and not __package__:
      from tools import run_in_terminal
    else:
      from .tools import run_in_terminal
    artistic_table = run_in_terminal.Run(filename, predicates)
    print(artistic_table)
    return

  program_text = open(filename).read()

  try:
    parsed_rules = parse.ParseFile(program_text,
                                   import_root=GetImportRoot())['rule']
  except parse.ParsingException as parsing_exception:
    parsing_exception.ShowMessage()
    sys.exit(1)

  if command == 'parse':
    # Minimal indentation for better readability of deep objects.
    print(json.dumps(parsed_rules, sort_keys=True, indent=' '))
    return 0

  if command == 'infer_types':
    typing_engine = infer.TypesInferenceEngine(parsed_rules)
    typing_engine.InferTypes()
    # print(parsed_rules)
    print(json.dumps(parsed_rules, sort_keys=True, indent=' '))
    return 0

  if command == 'show_signatures':
    try:
      logic_program = universe.LogicaProgram(parsed_rules)
      if not logic_program.typing_engine:
        logic_program.RunTypechecker()
    except infer.TypeErrorCaughtException as type_error_exception:
      print(logic_program.typing_engine.ShowPredicateTypes())
      type_error_exception.ShowMessage()
      return 1
    print(logic_program.typing_engine.ShowPredicateTypes())
    return 0

  predicates_list = predicates.split(',')

  user_flags = ReadUserFlags(parsed_rules, argv[4:])

  if command == 'build_schema':
    logic_program = universe.LogicaProgram(parsed_rules, user_flags=user_flags)
    engine = logic_program.annotations.Engine()
    type_retrieval_service = type_retrieval_service_discovery\
      .get_type_retrieval_service(engine, parsed_rules, predicates_list)
    type_retrieval_service.RetrieveTypes(filename)
    return 0

  for predicate in predicates_list:
    try:
      logic_program = universe.LogicaProgram(
          parsed_rules, user_flags=user_flags)
      formatted_sql = logic_program.FormattedPredicateSql(predicate)
      preamble = logic_program.execution.preamble
      defines_and_exports = logic_program.execution.defines_and_exports
      main_predicate_sql = logic_program.execution.main_predicate_sql
    except rule_translate.RuleCompileException as rule_compilation_exception:
      rule_compilation_exception.ShowMessage()
      sys.exit(1)
    except functors.FunctorError as functor_exception:
      functor_exception.ShowMessage()
      sys.exit(1)
    except infer.TypeErrorCaughtException as type_error_exception:
      type_error_exception.ShowMessage()
      sys.exit(1)
    except parse.ParsingException as parsing_exception:
      parsing_exception.ShowMessage()
      sys.exit(1)

    if command == 'print':
      print(formatted_sql)

    engine = logic_program.annotations.Engine()

    if command == 'run' or command == 'run_to_csv':
      # We should split and move this logic to dialects.
      if engine == 'bigquery':
        output_format = 'csv' if command == 'run_to_csv' else 'pretty'
        p = subprocess.Popen(['bq', 'query',
                              '--use_legacy_sql=false',
                              '--format=%s' % output_format],
                             stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        o, _ = p.communicate(formatted_sql.encode())
      elif engine == 'sqlite':
        # TODO: Make multi-statement scripts work.
        format = ('artistictable' if command == 'run' else 'csv')
        statements_to_execute = (
          [preamble] + defines_and_exports + [main_predicate_sql])
        o = sqlite3_logica.RunSqlScript(statements_to_execute,
                                        format).encode()
      elif engine == 'psql':
        p = subprocess.Popen(['psql', '--quiet'] +
                             (['--csv'] if command == 'run_to_csv' else []),
                             stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        commands = []
        o, _ = p.communicate(
            '\n'.join(commands + [formatted_sql]).encode())
      elif engine == 'trino':
        a = logic_program.annotations.annotations['@Engine']['trino']
        params = GetTrinoParameters(a)
        p = subprocess.Popen(['trino'] + params +
                             (['--output-format=CSV_HEADER_UNQUOTED']
                              if command == 'run_to_csv' else
                              ['--output-format=ALIGNED']),
                              stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        o, _ = p.communicate(formatted_sql.encode())
      elif engine == 'presto':
        a = logic_program.annotations.annotations['@Engine']['presto']
        catalog = a.get('catalog', 'memory')
        server = a.get('server', 'localhost:8080')
        p = subprocess.Popen(['presto',
                              '--catalog=%s' % catalog,
                              '--server=%s' % server,
                              '--file=/dev/stdin'] +
                             (['--output-format=CSV_HEADER_UNQUOTED']
                              if command == 'run_to_csv' else
                              ['--output-format=ALIGNED']),
                              stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        o, _ = p.communicate(formatted_sql.encode())
      else:
        assert False, 'Unknown engine: %s' % engine
      print(o.decode())


def run_main():
  """Run main function with system arguments."""
  main(sys.argv)


if __name__ == '__main__':
  main(sys.argv)



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Common utilities for Logica predicate compilation and execution."""

import subprocess
import sys

if '.' not in __package__:
  from common import sqlite3_logica
  from compiler import functors
  from compiler import rule_translate
  from compiler import universe
  from parser_py import parse
else:
  from ..common import sqlite3_logica
  from ..compiler import functors
  from ..compiler import rule_translate
  from ..compiler import universe
  from ..parser_py import parse


def ParseOrExit(filename, import_root=None):
  """Parse a Logica program."""
  with open(filename) as f:
    program_text = f.read()

  try:
    parsed_rules = parse.ParseFile(program_text,
                                   import_root=import_root)['rule']
  except parse.ParsingException as parsing_exception:
    parsing_exception.ShowMessage()
    sys.exit(1)

  return parsed_rules


def GetProgramOrExit(filename, user_flags=None, import_root=None):
  """Get program object from a file."""
  parsed_rules = ParseOrExit(filename, import_root=import_root)
  try:
    p = universe.LogicaProgram(parsed_rules, user_flags=user_flags)
  except rule_translate.RuleCompileException as rule_compilation_exception:
    rule_compilation_exception.ShowMessage()
    sys.exit(1)
  except functors.FunctorError as functor_exception:
    functor_exception.ShowMessage()
    sys.exit(1)
  return p


def RunQuery(sql,
             settings=None,
             output_format='pretty', engine='bigquery'):
  """Run a SQL query on BigQuery."""
  settings = settings or {}
  if engine == 'bigquery':
    p = subprocess.Popen(['bq', 'query',
                          '--use_legacy_sql=false',
                          '--format=%s' % output_format],
                         stdin=subprocess.PIPE, stdout=subprocess.PIPE)
  elif engine == 'sqlite':
    # TODO: Make multi-statement scripts work.
    return sqlite3_logica.RunSQL(sql)
  elif engine == 'psql':
    p = subprocess.Popen(['psql', '--quiet'],
                         stdin=subprocess.PIPE, stdout=subprocess.PIPE)
  elif engine == 'trino':
    catalog = settings.get('catalog', 'memory')
    server = settings.get('server', 'http://localhost:8080')
    p = subprocess.Popen(['trino',
                          '--catalog=%s' % catalog,
                          '--server=%s' % server] +
                          ['--output-format=ALIGNED'],
                          stdin=subprocess.PIPE, stdout=subprocess.PIPE)
  elif engine == 'presto':
    catalog = settings.get('catalog', 'memory')
    server = settings.get('server', 'http://localhost:8080')
    p = subprocess.Popen(['presto',
                          '--catalog=%s' % catalog,
                          '--server=%s' % server,
                          '--file=/dev/stdin'] +
                          ['--output-format=ALIGNED'],
                          stdin=subprocess.PIPE, stdout=subprocess.PIPE)
  else:
    assert False, 'Unknown engine: %s' % engine
  o, _ = p.communicate(sql.encode())
  return o.decode()


def RunPredicate(filename, predicate,
                 output_format='pretty', user_flags=None,
                 import_root=None):
  """Run a predicate on BigQuery."""
  p = GetProgramOrExit(filename, user_flags=user_flags,
                       import_root=import_root)
  sql = p.FormattedPredicateSql(predicate)
  engine = p.annotations.Engine()
  if ('@Engine' in p.annotations.annotations and
      engine in p.annotations.annotations['@Engine']):
    settings = p.annotations.annotations['@Engine'][engine]
  else:
    settings = {}
  return RunQuery(sql, settings,
                  output_format, engine=engine)


def RunQueryPandas(sql, engine, connection=None):
  """Running SQL query on the engine, returning Pandas dataframe."""
  import pandas
  if connection is None and engine == 'sqlite':
    connection = sqlite3_logica.SqliteConnect()
  if connection is None:
    assert False, 'Connection is required for engines other than SQLite.'
  if engine == 'bigquery':
    return connection.query(sql).to_dataframe()
  elif engine == 'psql':
    return pandas.read_sql(sql, connection)
  elif engine == 'sqlite':
    statements = parse.SplitRaw(sql, ';')[:-1]
    if len(statements) > 1:
      connection.executescript(';\n'.join(statements[:-1]))
    return pandas.read_sql(statements[-1], connection)
  else:
    raise Exception('Logica only supports BigQuery, PostgreSQL and SQLite '
                    'for now.')


def RunPredicateToPandas(filename, predicate,
                         user_flags=None, import_root=None, connection=None):
  p = GetProgramOrExit(filename, user_flags=user_flags,
                       import_root=import_root)
  sql = p.FormattedPredicateSql(predicate)
  engine = p.annotations.Engine()
  return RunQueryPandas(sql, engine, connection=connection)



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Platonic object checks whether a json-like Python object obeys schema.

Example:
  my_schema = PlatonicObject({"a": Num, "b": {"c": Str}})
  print(my_schema, {"a": 1, "b": {"c": "Hi!"}})
  # prints true, None.
  print(my_schema, {"a": "wow", "b": {"c": "Hi!", "d": "Bye!"}})
  # prints false for any of two reasons:
  #   type of a is incorrect,
  #   b has unexpected field.

Fields in records are never required, missing field is considered OK.
"""


class PlatonicAdherence:

  def __init__(self, the_object, success, error):
    self.the_object = the_object
    self.success = success
    self.error = error

  def __bool__(self):
    return not self.error
  
  def __str__(self):
    return self.error if self.error else 'OK'
  
  def Elements(self):
    return (self.the_object, self.success, self.error)


class PlatonicObject:
  """Platonic object which instances check schema."""
  def __init__(self, fields):
    self.fields = fields

  def __call__(self, x):
    if adherence := self.AdheredBy(x):
      return x
    assert False, 'Type error:' + str(adherence)

  def AdheredBy(self, record, path='',
               original_record=None) -> PlatonicAdherence:
    original_record = original_record or record  

    def MakeError(message):
      """Contextualized error message."""
      return 'Record %s, at path: %s encountered error: %s' % (
          original_record, path, message)

    if not isinstance(record, dict):
      return PlatonicAdherence(record, False, MakeError(
          'Expected record, but got not a dictionary.'))

    # Going over observed fields.
    for field_name, field_value in record.items():
      expected_type = self.fields.get(field_name, None)
      # If the field does not exist, then return error.
      if field_name not in self.fields:
        return PlatonicAdherence(
          record, False, MakeError('Unexpected field: %s' % field_name))
      
      if isinstance(expected_type, PlatonicObject):  # If field is an object.
        adherence = expected_type.AdheredBy(
            field_value, path=path + '.' + field_name,
            original_record=original_record)
        if not adherence:
          return adherence
      elif isinstance(expected_type, list):  # If field is a list.
        if not isinstance(field_value, list):
          return PlatonicAdherence(record, False,
                  MakeError(
                      'Field value %s was expected to be a list, got: %s.' % (
                          field_name, field_value)))
        [element_type] = expected_type
        for element in field_value:
          if isinstance(element_type, PlatonicObject):
            adherence = element_type.AdheredBy(
                element, path=path + '.' + field_name,
                original_record=original_record)
            if not adherence:
              return adherence
          else:
            assert False, MakeError('Unexpected element of list type: %s' % repr(element_type))
      else:
        assert False, MakeError('Bad type: %s' % expected_type)
    return PlatonicAdherence(record, True, None)

  def __str__(self):
    return '{%s}' % (
        ', '.join(
            '%s: %s' % (k, str(v))
            for k, v in self.fields.items()
        )
    )


  def __repr__(self):
    return self.__str__()


class Pod(PlatonicObject):
  """Platonic object for pods, i.e. numbers, strings, booleans."""
  def __init__(self, pod_type):
    self.pod_type = pod_type


  def AdheredBy(self, record, path='', original_record=None):
    original_record = original_record or record

    if isinstance(record, self.pod_type):
      return PlatonicAdherence(record, True, None)
    else:
      return PlatonicAdherence(
        record, False, '%s%s: Expected %s, got %s' % (
            original_record, path, self.pod_type.__name__, repr(record)))

  def __str__(self):
    return self.__class__.__name__ + '(' + str(self.pod_type.__name__) + ')'

  def __repr__(self):
    return self.__str__()


class NumericMeta:
  """Numeric metaclass.
  
  An instance of this class is Numeric type, usable as an argument of Pod.

  Numeric means integer or float.
  """
  def __instancecheck__(self, x):
    return isinstance(x, int) or isinstance(x, float)

  def __init__(self):
    self.__name__ = 'Num'


Int = Pod(int)
Str = Pod(str)
Float = Pod(float)
Bool = Pod(bool)
Num = Pod(NumericMeta())


#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, cast

from type_inference.type_inference_exception import TypeInferenceException
from type_inference.types.variable_types import AnyType, NumberType, StringType, ListType, RecordType, Type, AtomicType, \
  BoolType


def Rank(x):
  if isinstance(x, AnyType):
    return 0
  if isinstance(x, BoolType):
    return 1
  if isinstance(x, NumberType):
    return 2
  if isinstance(x, StringType):
    return 3
  if isinstance(x, AtomicType):
    return 4
  if isinstance(x, ListType):
    return 5
  if isinstance(x, RecordType):
    if x.is_opened:
      return 6
    return 7


def Intersect(a: Type, b: Type, bounds: Tuple[int, int]) -> Type:
  if Rank(a) > Rank(b):
    a, b = b, a

  if isinstance(a, AnyType):
    return b

  if isinstance(a, BoolType):
    if a == b:
      return a
    raise TypeInferenceException(a, b, bounds)

  if isinstance(a, AtomicType):
    if a == b or type(b) == AtomicType:
      return a
    raise TypeInferenceException(a, b, bounds)

  if isinstance(a, ListType):
    if isinstance(b, ListType):
      new_element = Intersect(a.element, b.element, bounds)
      return ListType(new_element)
    raise TypeInferenceException(a, b, bounds)

  a = cast(RecordType, a)
  b = cast(RecordType, b)
  if a.is_opened:
    if b.is_opened:
      return IntersectFriendlyRecords(a, b, True, bounds)
    else:
      if set(a.fields.keys()) <= set(b.fields.keys()):
        return IntersectFriendlyRecords(a, b, False, bounds)
      raise TypeInferenceException(a, b, bounds)
  else:
    if set(a.fields.keys()) == set(b.fields.keys()):
      return IntersectFriendlyRecords(a, b, False, bounds)
    raise TypeInferenceException(a, b, bounds)


def IntersectFriendlyRecords(a: RecordType, b: RecordType, is_opened: bool, bounds: Tuple[int, int]) -> RecordType:
  result = RecordType({}, is_opened)
  for name, f_type in b.fields.items():
    if name in a.fields:
      intersection = Intersect(f_type, a.fields[name], bounds)
      result.fields[name] = intersection
    else:
      result.fields[name] = f_type
  return result


def IntersectListElement(a_list: ListType, b_element: Type, bounds: Tuple[int, int]) -> Type:
  return Intersect(a_list.element, b_element, bounds)



#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Python representation of Logica propositions, expressions and rules.
# Example usage:
# P = avatar.Predicate('P')
# T = avatar.Predicate('T')
# F = avatar.Predicate('F')
# x, y, z = avatar.Variables('x', 'y', 'z')
# Aggr = avatar.Aggregation
# print(+P(a=x, b=Aggr('List', y), c = F(x=y)) << T(x, y, z=x))
#
# Prints:
# P(a: x, b? List= y, c: F(x: y)) distinct :- T(x, y, z: x)

class LogicalTerm:
  """Proposition or expression."""
  @classmethod
  def FromSyntax(cls, syntax):
    if 'call' in syntax:
      p = Predicate(syntax['call']['predicate_name'])
      positional_args = []
      named_args = {}
      for fv in syntax['call']['record']['field_value']:
        value = cls.FromSyntax(fv['value']['expression'])
        if isinstance(fv['field'], int):
          positional_args[fv['field']:fv['field']] = [value]
        else:
          named_args[fv['field']] = value
      return p(*positional_args, **named_args)

    if 'literal' in syntax:
      literal = syntax['literal']
      if 'the_number' in literal:
        number_str = literal['the_number']['number']
        if '.' in number_str:
          return Literal(float(number_str))
        return Literal(int(number_str))
      if 'the_string' in literal:
        return Literal(literal['the_string']['the_string'])
      if 'the_list' in literal:
        return Literal([LogicalTerm.FromSyntax(e)
                                for e in literal['the_list']['element']])
      assert False, syntax

    assert False, syntax


class Literal(LogicalTerm):
  def __init__(self, value):
    self.value = value
  
  def __str__(self):
    def Render(x):
      if isinstance(x, int) or isinstance(x, float):
        return str(x)
      if isinstance(x, list):
        return '[%s]' % ', '.join(map(Render, x))
      if isinstance(x, dict):
        return '{%s}' % ', '.join('%s: %s' % (k, Render(v))
                                  for k, v in x.items())
      if isinstance(x, str):
        assert '"' not in x, x
        assert '\\' not in x, x
        return '"%s"' % x
      if isinstance(x, LogicalTerm):
        return str(x)
      assert False, x
    return Render(self.value)

  def AsJson(self):
    value = self.value
    if (isinstance(value, int) or
        isinstance(value, float) or
        isinstance(value, str)):
      return value
    if isinstance(value, list):
      return [x.AsJson() for x in value]
    if isinstance(value, dict):
      return {k: v.AsJson() for k, v in value}
    assert False

class PredicateCall(LogicalTerm):
  def __init__(self, predicate_name,
               positional_args, named_args,
               distinct_denoted=None):
    def Digest(x):
      if (isinstance(x, int) or isinstance(x, float) or
          isinstance(x, list) or isinstance(x, dict)):
        return Literal(x)
      return x
    self.predicate_name = predicate_name
    self.positional_args = list(map(Digest, positional_args))
    self.named_args = {k: Digest(v) for k, v in named_args.items()}
    self.names_aggregated = [name
                             for name in self.named_args
                             if isinstance(self.named_args[name], Aggregation)]
    self.distinct_denoted = distinct_denoted

  def __pos__(self):
    return PredicateCall(self.predicate_name, self.positional_args, self.named_args,
                         True)

  def __str__(self):
    def RenderKeyValue(k, v):
      if k in self.names_aggregated:
        return '%s? %s= %s' % (
          k, v.aggregating_operator, v.aggregated_expression
        )
      if isinstance(v, Variable) and k == v.name:
        return '%s:' % k
      return '%s: %s' % (k, v)
    assert self.distinct_denoted or not self.names_aggregated, (
      self.predicate_name, self.positional_args, self.named_args,
      self.names_aggregated, self.distinct_denoted)

    for x in list(self.positional_args) + list(self.named_args.values()):
      assert isinstance(x, LogicalTerm), x

    positional = ', '.join(map(str, self.positional_args))
    named = ', '.join(RenderKeyValue(k, v)
                      for k, v in self.named_args.items())
    t = filter(None, [positional, named])
    distinct_str = ' distinct' if self.distinct_denoted else ''
    return '%s(%s)%s' % (self.predicate_name, ', '.join(t),
                         distinct_str)
  
  def __lshift__(self, body):
    return Rule(self, body)

  def __and__(self, other_conjunct):
    if isinstance(other_conjunct, Conjunction):
      return Conjunction([self] + other_conjunct.conjuncts)
    return Conjunction([self, other_conjunct])

  def __call__(self, *positional_args, **named_args):
    new_positional_args = self.positional_args
    new_positional_args[0:len(positional_args)] = positional_args
    new_named_args = dict(list(self.named_args.items()) +
                          list(named_args.items()))
    return PredicateCall(self.predicate_name,
                         new_positional_args, new_named_args)


class Subscript(LogicalTerm):
  def __init__(self, record, field):
    self.record = record
    self.field = field
  
  def __str__(self):
    return '%s.%s' % (self.record, self.field)

class Rule:
  def __init__(self, head: PredicateCall, body):
    self.head = head
    self.body = body
    self.comment_before_rule : str = None
  
  def __str__(self):
    if self.comment_before_rule:
      comment = '# ' + '# '.join(self.comment_before_rule.split('\n')) + '\n'
    else:
      comment = ''
    if self.body:
      return '%s%s :- %s' % (comment, self.head, self.body) 
    else:
      return comment + str(self.head)

class Conjunction(LogicalTerm):
  def __init__(self, conjuncts):
    self.conjuncts = conjuncts
  
  def __str__(self):
    return '\n  ' + ',\n  '.join(map(str, self.conjuncts))

  def __and__(self, other_conjunct):
    if isinstance(other_conjunct, Conjunction):
      return Conjunction(self.conjuncts + other_conjunct.conjuncts)
    return Conjunction(self.conjuncts + [other_conjunct])

class Disjunction(LogicalTerm):
  def __init__(self, disjuncts):
    self.disjuncts = disjuncts

  def __str__(self):
    return '\n  ' + ' |\n  '.join(map(lambda x: '(%s)' % x, self.disjuncts))

  def __or__(self, other_disjunct):
    if isinstance(other_disjunct, Disjunction):
      return Disjunction(self.disjuncts + other_disjunct.disjuncts)
    return Disjunction(self.disjuncts + [other_disjunct])


class Variable(LogicalTerm):
  def __init__(self, name):
    self.name = name
  
  def __str__(self):
    return self.name


def Variables(*variables):
  return map(Variable, variables)


class Aggregation(LogicalTerm):
  def __init__(self, aggregating_operator, aggregated_expression):
    self.aggregating_operator = aggregating_operator
    self.aggregated_expression = aggregated_expression

class Predicate:
  def __init__(self, predicate_name):
    self.predicate_name = predicate_name
  
  def __call__(self, *args, **named_args):
    return PredicateCall(self.predicate_name, args, named_args)

class Program:
  def __init__(self, rules):
    self.rules = rules
  
  def AddRule(self, rule):
    self.rules.append(rule)
  
  def __str__(self):
    return ';\n\n'.join(map(str, self.rules))
  


#!/usr/bin/python
#
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Utility to run pipeline in terminal with ASCII art showing progress.

import json
import os
import sys

if not __package__ or '.' not in __package__:
  from common import concertina_lib
  from compiler import universe
  from parser_py import parse
  from common import psql_logica
  from common import sqlite3_logica
  from compiler import functors
  from compiler import rule_translate
  from type_inference.research import infer
else:
  from ..common import concertina_lib
  from ..compiler import universe
  from ..parser_py import parse
  from ..common import psql_logica
  from ..common import sqlite3_logica
  from ..compiler import functors
  from ..compiler import rule_translate
  from ..type_inference.research import infer

class SqlRunner(object):
  def __init__(self, engine):
    self.engine = engine
    assert engine in ['sqlite', 'bigquery', 'psql']
    if engine == 'sqlite':
      self.connection = sqlite3_logica.SqliteConnect()
    else:
      self.connection = None
    if engine == 'bigquery':
      # To install:
      # python3 -m pip install google-auth
      from google import auth
      # To authenticate:
      # gcloud auth application-default login --project=${MY_PROJECT}
      credentials, project = auth.default()
    else:
      credentials, project = None, None
    if engine == 'psql':
      self.connection = psql_logica.ConnectToPostgres('environment')

    self.bq_credentials = credentials
    self.bq_project = project
  
  # TODO: Sqlite runner should not be accepting an engine.
  def __call__(self, sql, engine, is_final):
    return RunSQL(sql, engine, self.connection, is_final,
                  self.bq_credentials, self.bq_project)


def RunSQL(sql, engine, connection=None, is_final=False,
           bq_credentials=None, bq_project=None):
  if engine == 'bigquery':
    from google.cloud import bigquery
    client = bigquery.Client(credentials=bq_credentials,
                             project=bq_project)
    df = client.query(sql).to_dataframe()
    # Another way to query BQ:
    # import pandas
    # pandas.read_gbq(sql, project_id=bq_project_id)
    return list(df.columns), [list(r) for _, r in df.iterrows()]
  elif engine == 'psql':
    if is_final:
      cursor = psql_logica.PostgresExecute(sql, connection)
      rows = [list(map(psql_logica.DigestPsqlType, row))
              for row in cursor.fetchall()]
      return [d[0] for d in cursor.description], rows
    else:
      psql_logica.PostgresExecute(sql, connection)
  elif engine == 'sqlite':
    try:
      if is_final:
        cursor = connection.execute(sql)
        header = [d[0] for d in cursor.description]
        rows = cursor.fetchall()
        return header, rows
      else:
        connection.executescript(sql)
    except Exception as e:
      print("\n--- SQL ---")
      print(sql)
      print("Error while executing SQL:\n%s" % e)
      raise e
  else:
    raise Exception('Logica only supports BigQuery, PostgreSQL and SQLite '
                    'for now.')


def Run(filename, predicate_name, output_format='artistic_table'):
  try:
    rules = parse.ParseFile(open(filename).read())['rule']
  except parse.ParsingException as parsing_exception:
    parsing_exception.ShowMessage()
    sys.exit(1)


  try:
    program = universe.LogicaProgram(rules)
    engine = program.annotations.Engine()

    # This is needed to build the program execution.
    unused_sql = program.FormattedPredicateSql(predicate_name)

    (header, rows) = concertina_lib.ExecuteLogicaProgram(
        [program.execution], SqlRunner(engine), engine,
        display_mode='terminal')[predicate_name]
  except rule_translate.RuleCompileException as rule_compilation_exception:
    rule_compilation_exception.ShowMessage()
    sys.exit(1)
  except functors.FunctorError as functor_exception:
    functor_exception.ShowMessage()
    sys.exit(1)
  except infer.TypeErrorCaughtException as type_error_exception:
    type_error_exception.ShowMessage()
    sys.exit(1)

  if output_format == 'artistic_table':
    artistic_table = sqlite3_logica.ArtisticTable(header, rows)
    return artistic_table
  elif output_format == 'header_rows':
    return header, rows
  else:
    assert False, 'Unknown output format: %s' % output_format



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def GetRecursionFunctor(depth):
  """Returns functor that unfolds recursion.

  Example:
  P_r0 := P_recursive_head(P_recursive: nil);
  P_r1 := P_recursive_head(P_recursive: P_r0);
  P_r2 := P_recursive_head(P_recursive: P_r1);
  P_r3 := P_recursive_head(P_recursive: P_r2);
  P := P_r3();
  """
  result_lines = ['P_r0 := P_recursive_head(P_recursive: nil);']
  for i in range(depth):
      result_lines.append(
          'P_r{1} := P_recursive_head(P_recursive: P_r{0});'.format(i, i + 1))
  result_lines.append('P := P_r{0}();'.format(depth))
  return '\n'.join(result_lines)

def GetRenamingFunctor(member, root):
  """Renaming recursive cover member.

  Example.
  # Original:
  A(x) :- B(x);
  B(x) :- C(x);
  C(x) :- D(x);
  D(x) :- A(x);
  D(0);

  # After first renaming.
  D_recursive_head(x) :- A_recursive(x);
  D_recursive_head(0);

  C_recursive_head(x) :- D_recursive_head(x);

  B_recursive_head(x) :- C_recursive_head(x);

  A_recursive_head(x) :- B_recursive_head(x);

  A := A_r10();
  D := D_recursive_head(A_recursive: A);  # <-- renaming functor.
  C := C_recursive_head(A_recursive: A);
  B := B_recursive_head(A_recursive: A);
  """
  return '{0} := {0}_recursive_head({1}_recursive: {1});'.format(member, root)


#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3
from typing import Dict

from type_inference.inspectors.inspector_base import Inspector
from type_inference.inspectors.table_not_exist_exception import TableDoesNotExistException
from type_inference.types.variable_types import NumberType, StringType, Type


def Convert(sqlite_type: str) -> Type:
  lower_sqlite_type = sqlite_type.lower()
  if lower_sqlite_type == 'null':
    raise NotImplementedError()
  if lower_sqlite_type == 'integer' or lower_sqlite_type == 'real':
    return NumberType()
  if lower_sqlite_type == 'text':
    return StringType()
  raise NotImplementedError(sqlite_type)


class SQLiteInspector(Inspector):
  def __init__(self, db_path: str):
    self.db_path = db_path

  def TryGetColumnsInfo(self, table_name: str) -> Dict[str, Type]:
    with sqlite3.connect(self.db_path) as conn:
      cursor = conn.cursor()
      columns_info = cursor.execute(f'PRAGMA table_info({table_name});').fetchall()
      if len(columns_info) == 0:
        raise TableDoesNotExistException(table_name)
    return {column[1]: Convert(column[2]) for column in columns_info}



import setuptools

with open("logica/README.md", "r") as f:
  long_description = f.read()

setuptools.setup(
  name = "logica",
  version = "1.3.14159265",
  author = "Evgeny Skvortsov",
  author_email = "logica@evgeny.ninja",
  description = "Logica language.",
  long_description = long_description,
  long_description_content_type = "text/markdown",
  url="https://github.com/evgskv/logica",
  packages=setuptools.find_namespace_packages(),
  classifiers = [
      "Topic :: Database",
      "License :: OSI Approved :: Apache Software License"
  ],
  entry_points = {
    'console_scripts': ['logica=logica.logica:run_main']
  },
  python_requires= ">=3.0"
)




#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from logging import error

if '.' not in __package__:
  from common import color
  from type_inference import type_retrieval_exception
else:
  from ..common import color
  from ..type_inference import type_retrieval_exception


class UnknownBigQueryTypeException(type_retrieval_exception.TypeRetrievalException):
  def __init__(self, type: str):
    error(f'''{color.Format("[ {error}Error{end} ]")}'Unknown BigQuery type! Used type was {type}
          ''')



#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import psycopg2

if '.' not in __package__:
  from type_inference import type_retriever_base, psql_type_parser
else:
  from ..type_inference import type_retriever_base, psql_type_parser


class PostgresqlTypeRetriever(type_retriever_base.TypeRetrieverBase):
  def __init__(self):
    super().__init__()
    self.built_in_types = set()
    self.user_defined_types = dict()

  def ExtractTypeInfo(self, connection_string: str):
    if self.built_in_types and self.user_defined_types:
      return

    with psycopg2.connect(connection_string) as conn:
      with conn.cursor() as cur:
        # this SQL query takes all types and returns them
        # with boolean flag "is that type built in" and
        # JSON array of its fields (null for primitive types)
        cur.execute('''
SELECT t.typname                AS type,
       n.nspname = 'pg_catalog' AS is_built_in,
       CASE
           WHEN bool_and(pg_attribute.attname IS NULL) THEN NULL
           ELSE jsonb_agg(json_build_object('field_name', pg_attribute.attname, 'field_type', child_type.typname)) END
                                AS fields
FROM pg_type t
         LEFT JOIN pg_attribute ON pg_attribute.attrelid = t.typrelid
         LEFT JOIN pg_type AS child_type ON child_type.oid = pg_attribute.atttypid
         JOIN pg_catalog.pg_namespace n ON n.oid = t.typnamespace
WHERE n.nspname <> 'information_schema'
  AND (
            t.typrelid = 0 OR (SELECT c.relkind = 'c'
                               FROM pg_catalog.pg_class c
                               WHERE c.oid = t.typrelid)
    )
  AND NOT EXISTS(
        SELECT 1
        FROM pg_catalog.pg_type el
        WHERE el.oid = t.typelem
          AND el.typarray = t.oid
    )
GROUP BY t.typname, n.nspname;''')

        for type, is_built_in, fields in cur.fetchall():
          if is_built_in:
            self.built_in_types.add(type)
          else:
            self.user_defined_types[type] = {field['field_name']: field['field_type'] for field in fields} if fields else {}
  
  def UnpackType(self, type: str) -> str:
    if type.startswith('_'):
      return '[%s]' % self.UnpackTypeWithCaching(type[1:])
    
    if type in self.built_in_types:
      return psql_type_parser.PostgresTypeToLogicaType(type)
    
    fields = self.user_defined_types[type]
    fields = (f'{field_name}: {self.UnpackTypeWithCaching(field_type)}' for field_name, field_type in fields.items())
    return '{%s}' % (', '.join(fields))



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SQL dialects."""

import copy

if '.' not in __package__:
  from compiler.dialect_libraries import bq_library
  from compiler.dialect_libraries import psql_library
  from compiler.dialect_libraries import sqlite_library
  from compiler.dialect_libraries import trino_library
  from compiler.dialect_libraries import presto_library
  from compiler.dialect_libraries import databricks_library
else:
  from ..compiler.dialect_libraries import bq_library
  from ..compiler.dialect_libraries import psql_library
  from ..compiler.dialect_libraries import sqlite_library
  from ..compiler.dialect_libraries import trino_library
  from ..compiler.dialect_libraries import presto_library
  from ..compiler.dialect_libraries import databricks_library

def Get(engine):
  return DIALECTS[engine]()


class Dialect(object):
  pass

  # Default methods:
  def MaybeCascadingDeletionWord(self):
    return ''  # No CASCADE is needed by default.
  
  def PredicateLiteral(self, predicate_name):
    return "'predicate_name:%s'" % predicate_name


class BigQueryDialect(Dialect):
  """BigQuery SQL dialect."""

  def Name(self):
    return 'BigQuery'

  def BuiltInFunctions(self):
    return {}

  def InfixOperators(self):
    return {
        '++': 'CONCAT(%s, %s)',
    }

  def Subscript(self, record, subscript, record_is_table):
    return '%s.%s' % (record, subscript)

  def LibraryProgram(self):
    return bq_library.library

  def UnnestPhrase(self):
    return 'UNNEST({0}) as {1}'

  def ArrayPhrase(self):
    return 'ARRAY[%s]'

  def GroupBySpecBy(self):
    return 'name'

  def DecorateCombineRule(self, rule, var):
    return rule

  def PredicateLiteral(self, predicate_name):
    return 'STRUCT("%s" AS predicate_name)' % predicate_name

  
class SqLiteDialect(Dialect):
  """SqLite SQL dialect."""

  def Name(self):
    return 'SqLite'

  def BuiltInFunctions(self):
    return {
        'Set': 'DistinctListAgg({0})',
        'Element': "JSON_EXTRACT({0}, '$[' || {1} || ']')",
        'Range': ('(select json_group_array(n) from (with recursive t as'
                  '(select 0 as n union all '
                  'select n + 1 as n from t where n + 1 < {0}) '
                  'select n from t) where n < {0})'),
        'ValueOfUnnested': '{0}.value',
        'List': 'JSON_GROUP_ARRAY({0})',
        'Size': 'JSON_ARRAY_LENGTH({0})',
        'Join': 'JOIN_STRINGS({0}, {1})',
        'Count': 'COUNT(DISTINCT {0})',
        'StringAgg': 'GROUP_CONCAT(%s)',
        'Sort': 'SortList({0})',
        'MagicalEntangle': 'MagicalEntangle({0}, {1})',
        'Format': 'Printf(%s)',
        'Least': 'MIN(%s)',
        'Greatest': 'MAX(%s)',
        'ToString': 'CAST(%s AS TEXT)',
        'DateAddDay': "DATE({0}, {1} || ' days')",
        'DateDiffDay': "CAST(JULIANDAY({0}) - JULIANDAY({1}) AS INT64)"
    }

  def DecorateCombineRule(self, rule, var):
    return DecorateCombineRule(rule, var)

  def InfixOperators(self):
    return {
        '++': '(%s) || (%s)',
        '%' : '(%s) %% (%s)',
        'in': 'IN_LIST(%s, %s)'
    }

  def Subscript(self, record, subscript, record_is_table):
    if record_is_table:
      return '%s.%s' % (record, subscript)
    else:
      return 'JSON_EXTRACT(%s, "$.%s")' % (record, subscript)
  
  def LibraryProgram(self):
    return sqlite_library.library

  def UnnestPhrase(self):
    return 'JSON_EACH({0}) as {1}'

  def ArrayPhrase(self):
    return 'JSON_ARRAY(%s)'

  def GroupBySpecBy(self):
    return 'expr'

class PostgreSQL(Dialect):
  """PostgreSQL SQL dialect."""

  def Name(self):
    return 'PostgreSQL'

  def BuiltInFunctions(self):
    return {
        'Range': '(SELECT ARRAY_AGG(x) FROM GENERATE_SERIES(0, {0} - 1) as x)',
        'RangeOf' : '(SELECT ARRAY_AGG(x) FROM GENERATE_SERIES(0, ARRAY_LENGTH({0}, 1) - 1) as x)',
        'ToString': 'CAST(%s AS TEXT)',
        'ToInt64': 'CAST(%s AS BIGINT)',
        'Element': '({0})[{1} + 1]',
        'Size': 'COALESCE(ARRAY_LENGTH({0}, 1), 0)',
        'Count': 'COUNT(DISTINCT {0})',
        'MagicalEntangle': '(CASE WHEN {1} = 0 THEN {0} ELSE NULL END)',
        'ArrayConcat': '{0} || {1}',
        'Split': 'STRING_TO_ARRAY({0}, {1})'
      }

  def InfixOperators(self):
    return {
        '++': 'CONCAT(%s, %s)',
        'in': '%s = ANY(%s)'
    }

  def Subscript(self, record, subscript, record_is_table):
    return '(%s).%s' % (record, subscript)

  def LibraryProgram(self):
    return psql_library.library

  def UnnestPhrase(self):
    return 'UNNEST({0}) as {1}'

  def ArrayPhrase(self):
    return 'ARRAY[%s]'

  def GroupBySpecBy(self):
    return 'expr'

  def DecorateCombineRule(self, rule, var):
    return DecorateCombineRule(rule, var)
  
  def MaybeCascadingDeletionWord(self):
    return ' CASCADE'  # Need to cascade in PSQL.


class Trino(Dialect):
  """Trino analytic engine dialect."""

  def Name(self):
    return 'Trino'

  def BuiltInFunctions(self):
    return {
        'Range': 'SEQUENCE(0, %s - 1)',
        'ToString': 'CAST(%s AS VARCHAR)',
        'ToInt64': 'CAST(%s AS BIGINT)',
        'ToFloat64': 'CAST(%s AS DOUBLE)',
        'AnyValue': 'ARBITRARY(%s)',
        'ArrayConcat': '{0} || {1}'
    }

  def InfixOperators(self):
    return {
        '++': 'CONCAT(%s, %s)',
    }

  def Subscript(self, record, subscript, record_is_table):
    return '%s.%s' % (record, subscript)

  def LibraryProgram(self):
    return trino_library.library

  def UnnestPhrase(self):
    return 'UNNEST({0}) as pushkin({1})'

  def ArrayPhrase(self):
    return 'ARRAY[%s]'

  def GroupBySpecBy(self):
    return 'index'

  def DecorateCombineRule(self, rule, var):
    return rule


class Presto(Dialect):

  def Name(self):
    return 'Presto'

  def BuiltInFunctions(self):
    return {
        'Range': 'SEQUENCE(0, %s - 1)',
        'ToString': 'CAST(%s AS VARCHAR)',
        'ToInt64': 'CAST(%s AS BIGINT)',
        'ToFloat64': 'CAST(%s AS DOUBLE)',
        'AnyValue': 'ARBITRARY(%s)'
    }

  def InfixOperators(self):
    return {
        '++': 'CONCAT(%s, %s)',
    }

  def Subscript(self, record, subscript, record_is_table):
    return '%s.%s' % (record, subscript)

  def LibraryProgram(self):
    return presto_library.library

  def UnnestPhrase(self):
    return 'UNNEST({0}) as pushkin({1})'

  def ArrayPhrase(self):
    return 'ARRAY[%s]'

  def GroupBySpecBy(self):
    return 'index'

  def DecorateCombineRule(self, rule, var):
    return rule

def DecorateCombineRule(rule, var):
  """Resolving ambiguity of aggregation scope."""
  # Entangling result of aggregation with a variable that comes from a list
  # unnested inside a combine expression, to make it clear that aggregation
  # must be done in the combine. 
  rule = copy.deepcopy(rule)

  rule['head']['record']['field_value'][0]['value'][
    'aggregation']['expression']['call'][
    'record']['field_value'][0]['value'] = (
    {
      'expression': {
        'call': {
          'predicate_name': 'MagicalEntangle',
          'record': {
            'field_value': [
              {
                'field': 0,
                'value': rule['head']['record']['field_value'][0]['value'][
                  'aggregation']['expression']['call'][
                    'record']['field_value'][0]['value']      
              },
              {
                'field': 1,
                'value': {
                  'expression': {
                    'variable': {
                      'var_name': var
                    }
                  }
                }
              }
            ]
          }
        }
      }
    }
  )

  if 'body' not in rule:
    rule['body'] = {'conjunction': {'conjunct': []}}
  rule['body']['conjunction']['conjunct'].append(
    {
      "inclusion": {
        "list": {
          "literal": {
            "the_list": {
              "element": [
                {
                  "literal": {
                    "the_number": {
                      "number": "0"
                    }
                  }
                }
              ]
            }
          }
        },
        "element": {
          "variable": {
            "var_name": var
          }
        }
      }
    }      
  )
  return rule

class Databricks(Dialect):
    """Databricks dialect"""

    #TODO: add DATEDIFF and NOW function

    def Name(self):
        return 'Databricks'

    def BuiltInFunctions(self):
        return {
            'ToString': 'CAST(%s AS STRING)',
            'ToInt64': 'CAST(%s AS BIGINT)',
            'ToFloat64': 'CAST(%s AS DOUBLE)',
            'AnyValue': 'ANY_VALUE(%s)',
            'ILike': '({0}::string ILIKE {1})',
            'Like': '({0}::string LIKE {1})',
            'Replace': 'REPLACE({0}::string, {1}, {2})',
            'ArrayConcat': 'ARRAY_JOIN({0}, {1})',
            'JsonExtract': 'GET_JSON_OBJECT({0}, {1})',
            'JsonExtractScalar': 'GET_JSON_OBJECT({0}, {1})',
            'Length': 'ARRAY_SIZE(%s)',
            'DateDiff': 'DATEDIFF({0}, {1}, {2})',
            'IsNull': '({0} IS NULL)',
            'LogicalOr': 'BOOL_OR(%s)',
            'LogicalAnd': 'BOOL AND(%s)'
        }

    def InfixOperators(self):
        return {
            '++': 'CONCAT(%s, %s)',
            'in': 'ARRAY_CONTAINS(%s, %s)'
        }

    def Subscript(self, record, subscript):
        return '%s.%s' % (record, subscript)

    def LibraryProgram(self):
        return databricks_library.library

    def UnnestPhrase(self):
        return 'explode({0}) AS pushkin({1})'

    def ArrayPhrase(self):
        return 'ARRAY(%s)'

    def GroupBySpecBy(self):
        return 'index'

    def DecorateCombineRule(self, rule, var):
        return rule

DIALECTS = {
    'bigquery': BigQueryDialect,
    'sqlite': SqLiteDialect,
    'psql': PostgreSQL,
    'presto': Presto,
    'trino': Trino,
    'databricks': Databricks
}




#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


class TypeRetrieverBase(abc.ABC):
  """For all given types builds its string representation as composition of Logica's primitive types."""
  def __init__(self):
    self.name_to_type_cache = dict()

  def UnpackTypeWithCaching(self, type: str) -> str:
    if type not in self.name_to_type_cache:
      self.name_to_type_cache[type] = self.UnpackType(type)

    return self.name_to_type_cache[type]

  @abc.abstractmethod
  def UnpackType(self, type: str) -> str:
    """Returns string representation of the given type"""



#!/usr/bin/python
#
# Copyright 2023 Logica Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import cast

from type_inference.built_in_functions_types import built_in_concrete_types
from type_inference.inspectors.inspector_base import Inspector
from type_inference.intersection import Intersect, IntersectListElement
from type_inference.types.edge import Equality, EqualityOfElement, FieldBelonging, PredicateArgument
from type_inference.types.expression import PredicateAddressing
from type_inference.types.types_graph import TypesGraph
from type_inference.types.variable_types import AnyType, ListType, RecordType


class TypeInference:
  def __init__(self, graphs: dict, inspector: Inspector = None):
    self.inspector = inspector
    self.all_edges = []
    self.graphs = graphs

    for name, graph in graphs.items():
      self.all_edges.extend(map(lambda t: (name, t), graph.ToEdgesSet()))

    self.MergeGraphs()

  @staticmethod
  def GetEdgeContainingSameExpression(predicate_addressing: PredicateAddressing, graph: TypesGraph):
    for key, value in graph.expression_connections.items():
      if str(key) == str(predicate_addressing):
        return list(value.values())[0][0]

  @staticmethod
  def FindField(predicate_addressing: PredicateAddressing, graph: TypesGraph):
    edge = TypeInference.GetEdgeContainingSameExpression(predicate_addressing, graph)
    if edge.vertices[0] == predicate_addressing:
      return edge.vertices[0]
    else:
      return edge.vertices[1]

  def MergeGraphs(self):
    edges_to_add = []
    for predicate_name, graph in self.graphs.items():
      for p in graph.expression_connections:
        if isinstance(p, PredicateAddressing) and p.type == AnyType() and p.predicate_name != predicate_name:
          if p.predicate_name in self.graphs:
            to_link = self.FindField(p, self.graphs[p.predicate_name])
            edges_to_add.append(('no_graph', Equality(p, to_link, (-1, -1))))
          else:
            column_info = self.inspector.TryGetColumnsInfo(p.predicate_name)
            p.type = column_info[p.field]
    self.all_edges.extend(edges_to_add)

  def Infer(self):
    changed = True
    while changed:
      changed = False

      for graph_name, edge in self.all_edges:
        if isinstance(edge, Equality):
          edge = cast(Equality, edge)
          left, right = edge.left.type, edge.right.type
          result = Intersect(left, right, edge.bounds)
          if result != edge.left.type:
            edge.left.type = result
            changed = True
          if result != edge.right.type:
            edge.right.type = result
            changed = True
        elif isinstance(edge, EqualityOfElement):
          edge = cast(EqualityOfElement, edge)
          if isinstance(edge.list.type, AnyType):
            edge.list.type = ListType(AnyType())
          left, right = edge.element.type, cast(ListType, edge.list.type)
          result = IntersectListElement(right, left, edge.bounds)
          if result != edge.element.type:
            edge.element.type = result
            changed = True
          if ListType(result) != edge.list.type:
            edge.list.type = ListType(result)
            changed = True
        elif isinstance(edge, FieldBelonging):
          edge = cast(FieldBelonging, edge)
          if isinstance(edge.parent.type, AnyType):
            edge.parent.type = RecordType({}, True)
          record = cast(RecordType, edge.parent.type)
          field_name = edge.field.subscript_field
          if field_name in record.fields:
            result = Intersect(edge.field.type, record.fields[field_name], edge.bounds)
            if result != record.fields[field_name]:
              changed = True
              record.fields[field_name] = result
          else:
            changed = True
            record.fields[field_name] = edge.field.type
        elif isinstance(edge, PredicateArgument):
          predicate_name = edge.logica_value.predicate_name

          if predicate_name in built_in_concrete_types:
            arg_types, args = self.get_arguments(edge.logica_value, graph_name)
            correct_types = built_in_concrete_types[predicate_name](arg_types, edge.bounds)

            for field, type in correct_types.items():
              args[field].type = type

  def get_arguments(self, logica_value: PredicateAddressing, graph_name: str):
    arg_types = {}
    args = {}

    for arg, edges in self.graphs[graph_name].expression_connections[logica_value].items():
      if any(isinstance(e, PredicateArgument) for e in edges):
        arg_types[arg.field] = arg.type
        args[arg.field] = arg

    return arg_types, args
